/*
    Tucnak - VHF contest log
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>
    and authors of web browser Links 0.96

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/


#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#define __USE_GNU
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <stdarg.h>

#include "zstring.h"

#define mem_alloc malloc
#define mem_free free
#define mem_realloc realloc

/* error.c */
#define internal(args...) debug_internal(__FILE__, __LINE__, args)
/* #define dbg(args...) fprintf(stderr, args)*/
#define error(args...) fprintf(stderr, args)
#define STRERR(no) strerror_r(no, errbuf, sizeof(errbuf))
void dbg(char *m, ...);
void debug_internal(char *file, int line, char *s, ...);
void int_error( char *, ...);

/* main.c */
extern int opt_d, opt_h, opt_i, opt_s, opt_u, opt_v;
extern char *opt_hostname;

void init(void);
void terminate_all_subsystems(void);
char *parse_options(int argc, char **argv);
void usage(void);


/* select.c */
#define H_READ  0
#define H_WRITE 1
#define H_ERROR 2
#define H_DATA  3

typedef long ttime;

struct list_head {
    void *next;
    void *prev;
};

struct xlist_head {
    struct xlist_head *next;
    struct xlist_head *prev;
};

#define init_list(x) {(x).next=&(x); (x).prev=&(x);}
#define list_empty(x) ((x).next == &(x))
#define add_at_pos(p,x) do {(x)->next=(p)->next; (x)->prev=(p); (p)->next=(x); (x)->next->prev=(x);} while(0)
#define add_to_list(l,x) add_at_pos((struct xlist_head *)&(l),(struct xlist_head *)(x))
#define del_from_list(x) {((struct list_head *)(x)->next)->prev=(x)->prev; ((struct list_head *)(x)->prev)->next=(x)->next;}
#define foreach(e,l) for ((e)=(l).next; (e)!=(void *)&(l); (e)=(e)->next)

extern int terminate;

void select_loop(void (*init)(void));
#ifdef LEAK_DEBUG_LIST
void debug_set_handlers(char *file, int line, int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
#else
void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
#endif
void kill_timer(int id);
int install_timer(ttime t, void (*func)(void *), void * cba);


/* tcp.c */

#define TCP_PORT 10001
#define MAXRDBUF 1024
#define MAXWRBUF 1024
/* milliseconds */
#define CONNECTION_TIMEOUT 30000
#define SEND_ACK_TIMEOUT 10000
#define ACK_TIMEOUT 20000
#define REMOVE_TIMEOUT 10000

enum tcp_state { 
    TCP_INIT=0, 
    TCP_CONNECTING,
    TCP_CONNECTED,
    TCP_WAIT_ACK,
    TCP_DISCONNECTED,
    TCP_DEAD,  /* conn removed immediately */
};

struct conn {
    void *next;  
    void *prev;
    int sock; 
    enum tcp_state state;
    struct sockaddr_in sin;
    int relseq;
    int timer;
    char rdbuf[MAXRDBUF];
    int rdbuflen;
    char wrbuf[MAXWRBUF];
    int wrbuflen;
};

struct conn conns;  	   		/* list of server's connection, valid on server */
extern struct conn *gconn;      /* connection to client with grabbed control */
extern struct conn *sconn; 		/* connection to server, valid only on client */


int init_tcp(void);
void tcp_accept_handler(void *xxx);
void tcp_read_handler(void *xxx);
void tcp_exception_handler(void *xxx);
int tcp_send_state(struct conn *conn);
void tcp_disconnect(struct conn *conn, char *reason);
int conn_prod_state(struct conn *conn);
void tcp_set_state(struct conn *conn, enum tcp_state state);
char *tcp_state_s(enum tcp_state state);
void kill_conn_timer(void *xxx);
void send_ack_timer(void *xxx);
void remove_conn_timer(void *xxx);
void free_conn(struct conn *conn);
void tcp_connect(char *hostname, int port);
void tcp_connected_handler(void *xxx);
void after_connect(struct conn *conn);
void rel_write(struct conn *conn, char *s);
void rel_write_all(char *s);
void rel_read(struct conn *conn, char *line);
void rel_sread(struct conn *conn, char *cmd, zchar *zs);
void rel_cread(struct conn *conn, char *cmd, zchar *zs);


/* sdev.c */

#define SDEV_FILENAME "/dev/ttyS0"

struct sdev{
	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct conn *conn;
};


struct sdev *init_sdev(void);
int sdev_open(void);
int sdev_close(void);
char sd_chk(char *s, int len);

/* rc.c */

extern struct list_head cfg;

struct rcitem {
    void *next;  
    void *prev;
	char *key;
	char *value;
};

int read_rc_file(char *filename);
int read_rc_files(void);
char *rc_getopt(char *key);
void rc_info(void);


/* term.c */

#define KBD_ENTER   0x100
#define KBD_BS      0x101
#define KBD_TAB     0x102
#define KBD_ESC     0x103
#define KBD_LEFT    0x104
#define KBD_RIGHT   0x105
#define KBD_UP      0x106
#define KBD_DOWN    0x107
#define KBD_INS     0x108
#define KBD_DEL     0x109
#define KBD_HOME    0x10a
#define KBD_END     0x10b
#define KBD_PGUP    0x10c
#define KBD_PGDN    0x10d

#define KBD_F1      0x120
#define KBD_F2      0x121
#define KBD_F3      0x122
#define KBD_F4      0x123
#define KBD_F5      0x124
#define KBD_F6      0x125
#define KBD_F7      0x126
#define KBD_F8      0x127
#define KBD_F9      0x128
#define KBD_F10     0x129
#define KBD_F11     0x12a
#define KBD_F12     0x12b

#define KBD_CTRL_C  0x200

#define KBD_SHIFT   1
#define KBD_CTRL    2
#define KBD_ALT     4

#define ESC_TIMEOUT         200

struct term{
  	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct termios oldtermios;
	int timer;
};


extern struct term *term;

int init_term(void);
void free_term(void);
void term_read_handler(void *xxx);
void term_exception_handler(void *xxx);
int get_esc_code(char *str, int len, char *code, int *num, int *el);
int process_queue(void);
void kbd_timeout(void *);
int handle_key(int key, int shifts);

/* crig.c */

extern struct crig *crig;

struct crig{
	int ptt;
};

void init_crig(void);


/* rig.c */

extern struct rig *rig;

struct rig{
	int ptt;
};

void init_rig(void);


/* sdump.c */

#define MAXSDUMP 2

struct sdump{
	int fd[MAXSDUMP];
	char rdbuf[MAXRDBUF];
    int colors[MAXSDUMP];
};

struct sdump *init_sdump(void);
int sdump_open(int idx, char *filename);
void sdump_read_handler(void *xxx);

/* mike.c */

int mike_set(int compress, int flags, char *nat, int arobust, int asquelch, int asqdelay);
int mike_start(void);
int mike_stop(void);

/* speaker.c */

int speaker_set(int flags, int port, int ajitter, int ajitteridlet);
int speaker_start(void);
int speaker_stop(void);
