/*
    iprig - Remote radio via IP
    Copyright (C) 2007  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"
#include "speakfree.h"

int opt_d=1, opt_h=0, opt_i=0, opt_s=0, opt_v=0, opt_u=0;
char *opt_hostname=NULL;
static int ac;
static char **av;

void usage(){
    printf(
"\n"            
" iprig - Remote radio via IP networks\n"
" Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz\n"
"\n"
" Command line arguments:\n"
" -d debug mode\n"
" -h this help\n"
" -i informations about HW etc\n"
" -s server mode\n"
" -u dump mode (see sdump.c)\n"
" -v version\n"
"\n");
}
         
char *parse_options(int argc, char **argv){
    int c;

    while (1){
        c = getopt(argc, argv, "dhisuv");
        if (c == -1) break;

        switch (c){
            case 'd':
                opt_d++;
                break;

            case 'h':     
                opt_h++;
                break;

            case 'i':
                opt_i++;
                break;

            case 's':
                opt_s++;
                break;
           
            case 'v':
                opt_v++;
                break;

            case 'u':
                opt_u++;
                break;

            default:
                printf ("?? getopt returned character code 0%o ??\n", c);
        }
    }
    if (optind < argc){
        opt_hostname = argv[optind++];
    }

    if (optind < argc){
        printf ("non-option ARGV-elements: ");
        while (optind < argc) printf ("%s ", argv[optind++]);
        printf ("\n");
    }

    return NULL;
}
     
void init(){
    parse_options(ac, av);
    
    if (opt_v){
        printf ("iprig 0.1\nCopyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>\n");
        exit(0);
    }
    
    if (opt_h){
        usage();
        exit(0);
    }

    read_rc_files();
    if (opt_i){
        rc_info();
        exit(0);
    }
    
    if (opt_u){
        init_sdump();
        return;
    }
    
    if (opt_d){
        dbg("debug level %d\n", opt_d);
    }

	setpgrp();
    
    if (opt_s){
        /* server */
        init_rig();
        mike_set(fCompGSM, FLG_RTP, NULL, 1, 0, 12000);
        speaker_set(0, 0, 200000, 400000);
        init_tcp();
    }else{
        /* client */
		init_crig();
        mike_set(fCompGSM, FLG_RTP, NULL, 1, 0, 12000);
        speaker_set(0, 0, 200000, 400000);
        
        if (!opt_hostname){
            opt_hostname = rc_getopt("server");
            if (!opt_hostname){
                usage();
                exit(0);
            }
        }
		init_term();
        tcp_connect(opt_hostname, 10001);
    }
}

void terminate_all_subsystems(){
	dbg("\ncleanup\n");
	mike_stop();
	if (!opt_s){
		free_term();
	}
}

int main(int argc, char *argv[])
{
    ac = argc;
    av = (char **)argv;

    
    select_loop(init);
    terminate_all_subsystems();
    
    return 0;
}


