#define BYTE_SWAP_DEBUG
/* #define PGP_DEBUG */
/*

                        Speak Freely for Unix
                  Network sound transmission program

        Designed and implemented in July of 1991 by John Walker

*/

#include "speakfree.h"
#include "version.h"
#include "header.h"

#ifdef HAVE_DEV_RANDOM

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#endif
/*  Destination host descriptor.  */

struct destination {
    struct destination *dnext;        /* Next destination in list */
    char *server;                     /* Host name identifier string */
    struct sockaddr_in name;          /* Internet address */
    struct sockaddr_in ctrl;          /* RTCP port address */
#ifdef NAT_LAUNCH
    int dsock;                        /* Data socket file descriptor */
    int csock;                        /* Control socket file descriptor */
#endif
    char deskey[9];                   /* Destination DES key, if any */
    char rtpdeskey[9];                /* Destination RTP DES key, if any */
    char vatdeskey[9];                /* Destination VAT DES key, if any */
    char ideakey[17];                 /* Destination IDEA key, if any */
    char pgpkey[17];                  /* Destination PGP key, if any */
    char blowfish_spec;               /* Nonzero if Blowfish key specified */
    char aes_spec;                    /* Nonzero if AES key specified */
    char *otp;                        /* Key file */
};

static int sock;                      /* Communication socket */
#ifdef NAT_LAUNCH
static int xnat_socketsharing = FALSE;  /* Is NAT socket sharing in use ? */
static int xnat_consock, xnat_datasock; /* NAT socket FDs passed by sfspeaker */
#endif
static struct destination *dests = NULL, *dtail;
static int compressing = FALSE;       /* Compress sound: simple 2X */
static int gsmcompress = TRUE;        /* GSM compress buffers */
#ifdef BYTE_SWAP_DEBUG
static int gsm_byte_order_debug = FALSE; /* Send GSM length field with swapped bytes */
#endif
static int lpccompress = FALSE;       /* LPC compress buffers */
static int lpc10compress = FALSE;     /* LPC-10 compress buffers */
static int adpcmcompress = FALSE;     /* ADPCM compress buffers */
static int celpcompress = FALSE;      /* CELP compress buffers */
static int toasted = FALSE;           /* Sending already-compressed file ? */
static int robust = 1;                /* Robust duplicate packets mode */
static int rseq = 0;                  /* Robust mode packet sequence number */

#define DEFAULT_SQUELCH 4096          /* Default squelch level */
static int squelch = 0;               /* Squelch level if > 0 */
static int sqdelay = 12000;           /* Samples to delay before squelch */
static int sqwait = 0;                /* Squelch delay countdown */
#ifdef PUSH_TO_TALK
static int push = TRUE;               /* Push to talk mode */
static int talking = FALSE;           /* Push to talk button state */
//static int rawmode = FALSE;           /* Is terminal in raw mode ? */
#endif
static int ring = FALSE;              /* Force speaker & level on next pkt ? */
static int rtp = FALSE;               /* Use Internet Real-Time Protocol */
static int vat = FALSE;               /* Use VAT protocol */
static int agc = FALSE;               /* Automatic gain control active ? */
static int rgain = 33;                /* Current recording gain level */
static int spurt = TRUE;              /* True for first packet of talk spurt */
static int debugging = FALSE;         /* Debugging enabled here and there ? */
static FILE *audioDumpFile = NULL;    /* Audio dump file, if any */
static int havesound = FALSE;         /* True if we've acquired sound input */
static char hostname[20];             /* Host name to send with packets */
static int loopback = FALSE;          /* Remote loopback mode */
static gsm gsmh;                      /* GSM handle */
static struct adpcm_state adpcm = {0, 0}; /* ADPCM compression state */

static char *vatid = NULL;            /* VAT ID packet */
static int vatidl;                    /* VAT ID packet length */

static unsigned long ssrc;            /* RTP synchronisation source identifier */
static unsigned long timestamp;       /* RTP packet timestamp */
static unsigned short seq;            /* RTP packet sequence number */
static unsigned long rtpdesrand;      /* RTP DES random RTCP prefix */

static char *sdes = NULL;             /* RTP SDES packet */
static int sdesl;                     /* RTP SDES packet length */

static char curkey[9] = "";           /* Current DES key if curkey[0] != 0 */
static char currtpkey[9] = "";        /* Current RTP DES key if currtpkey[0] != 0 */
static char curvatkey[9] = "";        /* Current VAT DES key if currtpkey[0] != 0 */
static char curideakey[17] = "";      /* Current IDEA key if curideakey[0] != 0 */
static char curpgpkey[17] = "";       /* Current PGP key if curpgpkey[0] != 0 */
static char curblowfish_spec = FALSE; /* Nonzero if Blowfish key specified */
static char curaes_spec = FALSE;      /* Nonzero if AES key specified */
static char *curotp = NULL;           /* Key file buffer */
static int sound_packet;              /* Ideal samples/packet */
#ifdef HALF_DUPLEX
static struct in_addr localhost;      /* Our internet address */
static int halfDuplexMuted = FALSE;   /* Muted by half-duplex transmission */
#endif

#ifdef sgi
static long usticks;                  /* Microseconds per clock tick */
#endif

#define TimerStep   (7 * 1000000L)    /* Alarm interval in microseconds */

#ifdef HEWLETT_PACKARD
extern int HPTermHookChar;
#endif

/* Audio input and control device file names. */

#ifdef AUDIO_DEVICE_FILE
extern char *devAudioInput, *devAudioControl;
#endif

static int mike_pid = 0;
static int mike_pipe[2];

#define ucase(x)    (islower(x) ? toupper(x) : (x))

#ifndef USE_CURSES

#ifndef UNIX420
#define UNIX5
#endif

#ifdef UNIX420
#include <sgtty.h>
#endif /* UNIX420 */


/* Set raw mode on terminal file.  Basically, get the terminal into a
   mode in which all characters can be read as they are entered.  CBREAK
   mode is not sufficient.  */

#if 0
static void tty_rawmode(void)
{
#ifdef UNIX420
   struct sgttyb arg;

   ioctl(fileno(stdin), TIOCGETP, &arg);          /* get basic parameters */
   arg.sg_flags |= RAW;               /* set raw mode */
   arg.sg_flags &= ~ECHO;             /* clear echo mode */
   ioctl(fileno(stdin), TIOCSETP, &arg);          /* set it */
#endif /* UNIX420 */

#ifdef UNIX5
   struct termio term_params;

   ioctl(fileno(stdin), TCGETA, &old_term_params);
   term_params = old_term_params;
   term_params.c_iflag &= ~(ICRNL|IXON|IXOFF);  /* no cr translation */
   term_params.c_iflag &= ~(ISTRIP);   /* no stripping of high order bit */
   term_params.c_oflag &= ~(OPOST);    /* no output processing */       
   term_params.c_lflag &= ~(ISIG|ICANON|ECHO); /* raw mode */
   term_params.c_cc[4] = 1;  /* satisfy read after 1 char */
   ioctl(fileno(stdin), TCSETAF, &term_params);
#endif /* UNIX5 */
/*printf("\n(raw)\n");*/
}

/* Restore tty mode */

static void tty_normode(void)
{
#ifdef UNIX420
struct sgttyb arg;

   ioctl(fileno(stdin), TIOCGETP, &arg);          /* get basic parameters */
   arg.sg_flags &= ~RAW;              /* clear raw mode */
   arg.sg_flags |= ECHO;              /* set echo mode */
   ioctl(fileno(stdin), TIOCSETP, &arg);          /* set it */
#endif /* UNIX420 */

#ifdef UNIX5
   ioctl(fileno(stdin), TCSETAF, &old_term_params);
#endif /* UNIX5 */
/*printf("\n(cooked)\n");*/
}
#endif

#endif
/*  ULARM  --  Wrapper for setitimer() that looks like alarm()
               but accepts a time in microseconds.  */

static void ularm(long t)
{
    struct itimerval it;

    it.it_value.tv_sec = t / 1000000L;
    it.it_value.tv_usec = t % 1000000L;
    it.it_interval.tv_sec = it.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &it, NULL);
}

/*  GSMCOMP  --  Compress the contents of a sound buffer using GSM.  */

static void gsmcomp(struct soundbuf *sb)
{
    gsm_signal src[160];
    gsm_frame dst;
    int i, j, l = 0;
    char *dp = ((char *) sb->buffer.buffer_val) + sizeof(short);

    sb->compression |= fCompGSM;
    for (i = 0; i < sb->buffer.buffer_len; i += 160) {
        for (j = 0; j < 160; j++) {
            if ((i + j) < sb->buffer.buffer_len) {
                src[j] = audio_u2s(sb->buffer.buffer_val[i + j]);
            } else {
                src[j] = 0;
            }
        }
        gsm_encode(gsmh, src, dst);
        bcopy(dst, dp, sizeof dst);
        dp += sizeof dst;
        l += sizeof dst;
    }

    /* Hide original uncompressed buffer length in first 2 bytes of buffer. */
    
    *((short *) sb->buffer.buffer_val) = htons((short) sb->buffer.buffer_len);
#ifdef BYTE_SWAP_DEBUG
    if (gsm_byte_order_debug) {
        unsigned char *bsp = (unsigned char *) sb->buffer.buffer_val;

        bsp[0] = sb->buffer.buffer_len & 0xFF;
        bsp[1] = (sb->buffer.buffer_len >> 8) & 0xFF;
    }
#endif
    sb->buffer.buffer_len = l + sizeof(short);
}

/*  ADPCMCOMP  --  Compress the contents of a sound buffer using ADPCM.  */

static void adpcmcomp(struct soundbuf *sb)
{
    unsigned char *dp = (unsigned char *) sb->buffer.buffer_val;
    struct adpcm_state istate;

    istate = adpcm;
    sb->compression |= fCompADPCM;
    adpcm_coder_u(dp, (char *) dp, sb->buffer.buffer_len, &adpcm);
    sb->buffer.buffer_len /= 2;

    /* Hide the ADPCM encoder state at the end of this buffer.
       The shifting and anding makes the code byte-order
       insensitive. */

    dp += sb->buffer.buffer_len;
    *dp++ = ((unsigned int) istate.valprev) >> 8;
    *dp++ = istate.valprev & 0xFF;
    *dp = istate.index;
    sb->buffer.buffer_len += 3;
}

#ifdef ENABLE_LPC
/*  LPCCOMP  --  Compress the contents of a sound buffer using LPC.  */

static void lpccomp(struct soundbuf *sb)
{
    int i, l = 0;
    char *dp = ((char *) sb->buffer.buffer_val) + sizeof(short);
    unsigned char *src = ((unsigned char *) sb->buffer.buffer_val);
    lpcparams_t lp;

    sb->compression |= fCompLPC;
    for (i = 0; i < sb->buffer.buffer_len; i += LPC_FRAME_SIZE) {
        lpc_analyze(src + i, &lp);
        bcopy(&lp, dp, LPCRECSIZE);
        dp += LPCRECSIZE;
        l += LPCRECSIZE;
    }

    /* Hide original uncompressed buffer length in first 2 bytes of buffer. */

    *((short *) sb->buffer.buffer_val) = htons((short) sb->buffer.buffer_len);
    sb->buffer.buffer_len = l + sizeof(short);
}

/*  LPC10COMP  --  Compress the contents of a sound buffer using LPC-10.  */

static void lpc10comp(struct soundbuf *sb)
{
    unsigned char *dp = (unsigned char *) sb->buffer.buffer_val;
    unsigned char *src = ((unsigned char *) sb->buffer.buffer_val);

    sb->compression |= fCompLPC10;
    sb->buffer.buffer_len = lpc10encode(src, dp, sb->buffer.buffer_len);
}

/*  LPC10STUFF  --  Stuff last 16 bytes of LPC10 packet in sendinghost. */

static int lpc10stuff(struct soundbuf *sb, int pktlen)
{
    if ((sb->compression & fCompLPC10) && (sb->buffer.buffer_len > 16)) {
        bcopy(((char *) sb) + (pktlen - (sizeof sb->sendinghost)),
            sb->sendinghost, sizeof sb->sendinghost);
        pktlen -= sizeof sb->sendinghost;
    }
    return pktlen;
}

/*  CELPCOMP  --  Compress the contents of a sound buffer using CELP.  */

static void celpcomp(struct soundbuf *sb)
{
    short src[240];
    char celp_frame[18];
    int i, j, l = 0;
    char *dp = ((char *) sb->buffer.buffer_val) + sizeof(short);

    sb->compression |= fCompCELP;
    for (i = 0; i < sb->buffer.buffer_len; i += 240) {
        for (j = 0; j < 240; j++) {
            if ((i + j) < sb->buffer.buffer_len) {
                src[j] = audio_u2s(sb->buffer.buffer_val[i + j]);
            } else {
                src[j] = 0;
            }
        }
        celp_encode(src, celp_frame);
        bcopy(celp_frame, dp, sizeof celp_frame);
        dp += sizeof celp_frame;
        l += sizeof celp_frame;
    }

    /* Hide original uncompressed buffer length in first 2 bytes of buffer. */
    
    *((short *) sb->buffer.buffer_val) = htons((short) sb->buffer.buffer_len);
    sb->buffer.buffer_len = l + sizeof(short);
}

/*  CELPSTUFF   --  Stuff last 16 bytes of CELP packet in sendinghost. */

static int celpstuff(struct soundbuf *sb, int pktlen)
{
    if ((sb->compression & fCompCELP) && (sb->buffer.buffer_len > 16)) {
        bcopy(((char *) sb) + (pktlen - (sizeof sb->sendinghost)),
            sb->sendinghost, sizeof sb->sendinghost);
        pktlen -= sizeof sb->sendinghost;
    }
    return pktlen;
}

#endif

/*  SENDRTPCTRL  --  Send RTP or VAT control message, encrypting if
                     necessary.  */

static int sendrtpctrl(struct destination *d, struct sockaddr *destaddr,
                       char *msg, int msgl, int rtppkt)
{
//	dbg("sending control packet\n");
    return sendto(
#ifdef NAT_LAUNCH
                  d->csock,
#else
                  sock,
#endif
                   msg, msgl, 0, destaddr, sizeof d->ctrl);
}


void dumpsin(char *text, struct sockaddr_in *sin){
    long naddr;
	
	naddr = ntohl(sin->sin_addr.s_addr);
    dbg("%s %d.%d.%d.%d:%d\n", text, (naddr>>24)&0xff, (naddr>>16)&0xff, (naddr>>8)&0xff, naddr&0xff, ntohs(sin->sin_port));
	
}

/*  ADDEST  --  Add destination host to host list.  */

static int addest(struct conn *conn)
{
    struct destination *d;
   // struct hostent *h;
   // char *mcs;
#ifdef NAT_LAUNCH
    int dataport = Internet_Port;           /* Default data port number */
    int ctrlport = Internet_Port + 1;       /* Default control port number */
#else
    int curport = Internet_Port;
#endif

	
#if 0	
    /* If a multicast scope descriptor appears in the name, scan
       it and lop it off the end.  We'll apply it later if we discover
       this is actually a multicast address. */

    if ((mcs = strrchr(host, '/')) != NULL &&
        ((strchr(host, ':') != NULL) || (mcs != strchr(host, '/')))) {
        *mcs = 0;
    }

    /* If a port number appears in the name, scan it and lop
       it off.  We allow a slash as well as the documented colon
       as the delimiter to avoid confusing users familiar with
       VAT. */

    if ((mcs = strrchr(host, ':')) != NULL ||
        (mcs = strrchr(host, '/')) != NULL) {
        *mcs++ = 0;
#ifdef NAT_LAUNCH
        dataport = atoi(mcs);
        /* The control port can be specified independently of
           the data port, separated by a comma. */
        if ((mcs = strrchr(host, ',')) != NULL) {
            ctrlport = atoi(mcs);
        } else {
            ctrlport = dataport + 1;
        }
#else
        curport = atoi(mcs);
#endif
    }
#endif
	
    /* If it's a valid IP number, use it.  Otherwise try to look
       up as a host name. */
#if 0
    if ((naddr = inet_addr(host)) != -1) {
    } else {
        h = gethostbyname(host);
        if (h == 0) {
            fprintf(stderr, "%s: unknown host\n", host);
            return FALSE;
        }
        bcopy((char *) h->h_addr, (char *) &naddr, sizeof naddr);
    }
#endif	


    d = (struct destination *) malloc(sizeof(struct destination));
    d->dnext = NULL;
//    d->server = host;
    bcopy((char *) &(conn->sin.sin_addr), (char *) &(d->name.sin_addr), sizeof(d->name.sin_addr));
    bcopy((char *) &(conn->sin.sin_addr), (char *) &(d->ctrl.sin_addr), sizeof(d->ctrl.sin_addr));
    bcopy(curkey, d->deskey, 9);
    bcopy(currtpkey, d->rtpdeskey, 9);
    bcopy(curvatkey, d->vatdeskey, 9);
    bcopy(curideakey, d->ideakey, 17);
    bcopy(curpgpkey, d->pgpkey, 17);
    d->blowfish_spec = curblowfish_spec;
    d->aes_spec = curaes_spec;
    d->otp = curotp;
    d->name.sin_family = AF_INET;
#ifdef NAT_LAUNCH
    d->name.sin_port = htons(dataport);
#else
    d->name.sin_port = htons(curport);
#endif
    d->ctrl.sin_family = AF_INET;
#ifdef NAT_LAUNCH
    d->ctrl.sin_port = htons(ctrlport);
#else
    d->ctrl.sin_port = htons(curport + 1);
#endif
	dumpsin("adddest name=", &d->name);
	dumpsin("adddest ctrl=", &d->ctrl);
    if (dests == NULL) {
        dests = d;
    } else {
        dtail->dnext = d;
    }
    dtail = d;
#ifdef NAT_LAUNCH
    d->dsock = (xnat_socketsharing ? xnat_datasock : sock);
    d->csock = (xnat_socketsharing ? xnat_consock : sock);
#endif

    /* Send initial status and identity message in
       the correct protocol. */

    if (rtp) {
        if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                        sdes, sdesl, TRUE) < 0) {
            perror("sending initial RTCP SDES packet");
			dbg("sock %d\n", d->csock);
            return FALSE;
        }
    } else if (vat) {
        if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                       vatid, vatidl, FALSE) < 0) { 
            perror("sending initial VAT ID control packet");
            return FALSE;
        }
    } else {
        char pid = sdes[0];

        /* Set Speak Freely protocol flag in packet */
        sdes[0] = (sdes[0] & 0x3F) | (1 << 6);
        if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                        sdes, sdesl, TRUE) < 0) {
            perror("sending initial Speak Freely SDES control packet");
            return FALSE;
        }
        sdes[0] = pid;
    }

    return TRUE;
}

/*  TIMERTICK  --  Timer tick signal-catching function.  */

static void timertick()
{
    struct destination *d;

    for (d = dests; d != NULL; d = d->dnext) {
        if (rtp) {
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                         sdes, sdesl, TRUE) < 0) {
                perror("sending RTCP SDES packet");
            }
        } else if (vat) {
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                        vatid, vatidl, FALSE) < 0) { 
                perror("sending VAT ID control packet");
            }
        } else {
            char pid = sdes[0];

            /* Set Speak Freely protocol flag in packet */
            sdes[0] = (sdes[0] & 0x3F) | (1 << 6);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                            sdes, sdesl, TRUE) < 0) {
                perror("sending Speak Freely SDES control packet");
            }
            sdes[0] = pid;
        }
    }

    ularm(TimerStep);
    signal(SIGALRM, timertick);       /* Reset signal to handle timeout */
}

/*  SENDPKT  --  Send a message to all active destinations.  */

static int sendpkt(struct soundbuf *sb)
{
    struct destination *d;
    int pktlen;

    if (gsmcompress && !toasted) {
        gsmcomp(sb);
    }

    if (adpcmcompress && !toasted) {
        adpcmcomp(sb);
    }

#ifdef ENABLE_LPC    
    if (lpccompress && !toasted) {
        lpccomp(sb);
    }

    if (lpc10compress && !toasted) {
        lpc10comp(sb);
    }
    
    if (celpcompress && !toasted) {
        celpcomp(sb);
    }
#endif    

    pktlen = sb->buffer.buffer_len + (sizeof(struct soundbuf) - BUFL);

    if (vat) {
        pktlen = vatout(sb, 0L, timestamp, spurt);
        timestamp += sound_packet;
    }

    if (rtp) {
        pktlen = rtpout(sb, ssrc, timestamp, seq, spurt);
        seq++;
        timestamp += sound_packet;
    }

    spurt = FALSE;                    /* Not the start of a talk spurt */

    for (d = dests; d != NULL; d = d->dnext) {

        {
            int i, wasf = FALSE;

            if (!vat && !rtp) {
                wasf = TRUE;
#ifdef HAVE_LPC                
                if (lpc10compress && !toasted) {
                    pktlen = lpc10stuff(sb, pktlen);
                }
                if (celpcompress && !toasted) {
                    pktlen = celpstuff(sb, pktlen);
                }
#endif                
                if (robust > 1) {
                    sb->compression |= fCompRobust;
                    sb->buffer.buffer_len |= ((long) rseq) << 24;
                    rseq = (rseq + 1) & 0xFF;
                }
                sb->compression = htonl(sb->compression);
                sb->buffer.buffer_len = htonl(sb->buffer.buffer_len);
            }

            for (i = 0; i < robust; i++) {
                if (sendto(
#ifdef NAT_LAUNCH
                    d->dsock,
#else           
                    sock,
#endif
                    (char *) sb, pktlen,
                    0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
                    perror("sending unencrypted audio");
                    return FALSE;
                }
            }
            if (wasf) {
                sb->compression = ntohl(sb->compression);
                sb->buffer.buffer_len = ntohl(sb->buffer.buffer_len);
                if (sb->compression & fCompRobust) {
                    sb->compression &= ~fCompRobust;
                    sb->buffer.buffer_len &= 0xFFFFFFL;
                }
                if ((sb->compression & fCompLPC10) &&
                    (sb->buffer.buffer_len > 16)) {
                    pktlen += sizeof(sb->sendinghost);
                }
                /* NOTE--If you implement support for CELP in RTP, you'll need to add
                   similar un-stuffing of the packet for CELP here. */
            }
        }
    }
    return TRUE;
}

/*  GETAUDIO  --  Open audio input.  If audio hardware is half duplex,
                  we may have to mute output in progress.  */

static int getaudio(void)
{
    int i;

#ifdef HALF_DUPLEX
    struct soundbuf hdreq;
    struct sockaddr_in name;

    /* If we're afflicted with half-duplex hardware, make several
       attempts to mute the output device and open audio input.
       We send a mute even if the output isn't active as a courtesy
       to the output program; if a later attempt to acquire output
       fails, he'll know it was a result of being muted. */

    for (i = 5; i > 0; i--) {
        hdreq.compression = htonl(fProtocol | fHalfDuplex | fHalfDuplexMute);
        strcpy(hdreq.sendinghost, hostname);
        hdreq.buffer.buffer_len = 0;

        bcopy((char *) &localhost, (char *) &(name.sin_addr), sizeof localhost);
        name.sin_family = AF_INET;
        name.sin_port = htons(Internet_Port);
        if (debugging) {
            fprintf(stderr,"Sending fHalfDuplex | fHalfDuplexMute Request\n");
        }
        if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL),
                   0, (struct sockaddr *) &name, sizeof name) < 0) {
            perror("sending half-duplex mute request");
        }
        sf_usleep(50000L);
        if (soundinit(O_RDONLY)) {
            halfDuplexMuted = TRUE;
            break;
        }
    }

    /* If we failed to initialise, send a resume just in case
       one of our mute requests made it through anyway. */

    if (i <= 0) {
        hdreq.compression = htonl(fProtocol | fHalfDuplex | fHalfDuplexResume);
        if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL),
                   0, (struct sockaddr *) &name, sizeof name) < 0) {
            perror("sending half-duplex resume request");
        }
    }
#else
    i = soundinit(O_RDONLY);
#endif

    if (i <= 0) {
        fprintf(stderr, "iprig: unable to initialise audio.\n");
        return FALSE;
    }
    havesound = TRUE;
    if (agc) {
        soundrecgain(rgain);      /* Set initial record level */
    }
    return TRUE;
}

/*  FREEAUDIO  --  Release sound input device, sending a resume to
                   the output device if we've muted it due to
                   half-duplex hardware.  */

static void freeaudio(void)
{
    if (havesound) {
        if (debugging) {
            fprintf(stderr, "Restoring audio modes at release.\n");
        }
        soundterm();
        havesound = FALSE;
    }

#ifdef HALF_DUPLEX

    /* When bailing out, make sure we don't leave the output
       muted. */

    if (halfDuplexMuted) {
        struct soundbuf hdreq;
        struct sockaddr_in name;

        hdreq.compression = htonl(fProtocol | fHalfDuplex | fHalfDuplexResume);
        strcpy(hdreq.sendinghost, hostname);
        hdreq.buffer.buffer_len = 0;

        bcopy((char *) &localhost, (char *) &(name.sin_addr), sizeof localhost);
        name.sin_family = AF_INET;
        name.sin_port = htons(Internet_Port);
        if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL),
                   0, (struct sockaddr *) &name, sizeof name) < 0) {
            perror("sending half-duplex resume request");
        }
        halfDuplexMuted = FALSE;
    }
#endif
}

/*  EXITING  --  Catch as many program termination signals as
                 possible and restore initial audio modes before
                 we exit.  */

static void exiting()
{
    struct destination *d;

    freeaudio();

    if (audioDumpFile != NULL) {
        fclose(audioDumpFile);
    }


    if (rtp) {
        for (d = dests; d != NULL; d = d->dnext) {
            char v[1024];
            int l;

            l = rtp_make_bye((unsigned char *) v, ssrc, "Exiting Speak Freely", TRUE);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                         v, l, TRUE) < 0) {
                perror("sending RTCP BYE packet");
            }
        }
    } else if (vat) {
        for (d = dests; d != NULL; d = d->dnext) {
            char v[16];
            int l;

            l = makevatdone(v, 0L);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                         v, l, FALSE) < 0) {
                perror("sending VAT DONE control packet");
            }
        }
    } else {
        for (d = dests; d != NULL; d = d->dnext) {
            char v[1024];
            int l;

            l = rtp_make_bye((unsigned char *) v, ssrc, "Exiting Speak Freely", TRUE);
            v[0] = (v[0] & 0x3F) | (1 << 6);
            if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
                         v, l, TRUE) < 0) {
                perror("sending Speak Freely BYE packet");
            }
        }
    }

    exit(0);
}

/*  CHANGEMODE  --  Reset state when talk/pause/quiet mode changes.  */

static void changemode(void)
{
    if (talking) {
        spurt = TRUE;
        /* Discard all backlog sound input. */
        soundflush();
    } else {
        sqwait = 0;                   /* Reset squelch delay */
    }
}


/*  SENDFILE  --  Send a file or, if the file name is NULL or a
                  single period, send real-time sound input. */

static int sendfile(char *f)
{ 
    soundbuf netbuf;
#define buf netbuf.buffer.buffer_val
    int /*bread, nread,*/ lread;
//    FILE *afile = NULL;
    static int firstTime = TRUE;

    if (firstTime) {
        firstTime = FALSE;
        signal(SIGHUP, exiting);     /* Set signal to handle termination */
        signal(SIGINT, exiting);     /* Set signal to handle termination */
        signal(SIGTERM, exiting);    /* Set signal to handle termination */
        signal(SIGALRM, timertick);  /* Set signal to handle timer */
        ularm(TimerStep);
    }

    /* Compute the number of sound samples needed to fill a
       packet of TINY_PACKETS bytes. */

    if (rtp) {
        sound_packet = (gsmcompress | lpccompress | adpcmcompress) ? (160 * 4)
                       : 320;
    } else if (vat) {
        sound_packet = (gsmcompress | lpccompress | adpcmcompress) ? (160 * 4)
                       : 320;
    } else {
        sound_packet = ((TINY_PACKETS - ((sizeof(soundbuf) - BUFL))) *
                        (compressing ? 2 : 1));
        if (gsmcompress) {
            sound_packet = compressing ? 3200 : 1600;
        } else if (adpcmcompress) {
            sound_packet *= 2;
            sound_packet -= 4;            /* Leave room for state at the end */
        } else if (lpccompress) {
            sound_packet = (compressing ? 2 : 1) * (10 * LPC_FRAME_SIZE);
        } else if (lpc10compress) {
            sound_packet = compressing ? 3600 : 1800;
        } else if (celpcompress) {
            sound_packet = compressing ? 3840 : 1920;
        }
    }

#ifdef SHOW_PACKET_SIZE
    printf("Samples per packet = %d\n", sound_packet);
#endif
    lread = sound_packet;

    strcpy(netbuf.sendinghost, hostname);

	{
            /* Send real-time audio. */

#ifdef PUSH_TO_TALK
/*        if (push) {
            char c;
            int l;

            fprintf(stderr,
                "Space bar switches talk/pause, Esc or \"q\" to quit, \".\" for chat text\nPause: ");
            fflush(stderr);
            rawmode = TRUE;
            while (TRUE) {
                while (TRUE) {
                    if (((l = read(fileno(stdin), &c, 1)) == 1) || (errno != EINTR)) {
                        break;
                    }
                }   
                if (l == 1) {
                    break;
                }
            }
            if (l != 1) {
                perror("waiting for first Pause/Talk character");
            }
            termchar(c);
            fprintf(stderr, squelch > 0 ? "\rQuiet:  " : "\rTalk:   ");
            fcntl(fileno(stdin), F_SETFL, O_NDELAY);
            talking = TRUE;
            spurt = TRUE;
        }    */
#endif
        fcntl(fileno(stdin), F_SETFL, O_NDELAY);
        talking = TRUE;
        spurt = TRUE;

        /* Send real-time sound. */

#ifdef Solaris
        setaubufsize(sound_packet);
#endif

        if (!getaudio()) {
            fprintf(stderr, "Unable to initialise audio.\n");
            return 2;
        }
        changemode();
        while (TRUE) {
            int soundel = 0;
            unsigned char *bs = (unsigned char *) buf;

            if (havesound) {

                /* Obtain a packet's worth of real-time audio from
                   the audio input hardware. */

                soundel = soundgrab(buf, sound_packet);

                /* If a raw audio dump file (-w option) is specified,
                   dump whatever we got from the audio input to the
                   designated file.  This can come in handy when
                   debugging audio drivers which don't provide samples
                   in the expected 8 bit mu-law mode. */

                if ((audioDumpFile != NULL) && (soundel > 0)) {
                    fwrite(buf, 1, soundel, audioDumpFile);
                }
            }

#ifdef PUSH_TO_TALK
            while (push) {
                char c;
                int rlen;

                if ((rlen = read(fileno(stdin), &c, 1)) > 0) {
					switch(c){
						case 'q':
							exiting();
							break;
							
						case '0':
							if (!talking) break;
							talking= FALSE;
#ifdef HALF_DUPLEX
	        	            /* For half-duplex, acquire and release the
    	        	           audio device at each transition.  This lets
        	        	       us mute the output only while in Talk mode. */
                        	freeaudio();
#endif
							break;
							
						case '1':
							if (talking) break;
							talking=TRUE;	
#ifdef HALF_DUPLEX
							if (!getaudio()) {
								fprintf(stderr, "Audio device busy.\n");
								talking = FALSE;
							}
#endif
							break;
					}
                    fprintf(stderr, talking ? (squelch > 0 ? "Quiet\n" : "Talk\n") : "Pause\n");
                    fcntl(0, F_SETFL, talking ? O_NDELAY : 0);
                    changemode();
                    break;
                } else {
                    if (rlen == -1 && errno == EINTR) {
                        continue;
                    }
                    break;
                }
            }
#endif

            if (soundel > 0) {
                register unsigned char *start = bs;
                register int j;
                int squelched = (squelch > 0), osl = soundel;

                /* If entire buffer is less than squelch, ditch it.  If
                   we haven't received sqdelay samples since the last
                   squelch event, continue to transmit. */

                if (sqdelay > 0 && sqwait > 0) {
                    if (debugging) {
                        printf("Squelch countdown: %d samples left.\n",
                            sqwait);
                    }
                    sqwait -= soundel;
                    squelched = FALSE;
#ifdef PUSH_TO_TALK
                    if (sqwait <= 0 && push) {
                        fprintf(stderr, "Quiet (2)\n");
                    }
#endif
                } else if (squelch > 0) {
                    for (j = 0; j < soundel; j++) {
#ifdef USQUELCH
                        if (((*start++ & 0x7F) ^ 0x7F) > squelch)
#else
                        int samp = audio_u2s(*start++);

                        if (samp < 0) {
                            samp = -samp;
                        }
                        if (samp > squelch)
#endif
                        {
                            squelched = FALSE;
#ifdef PUSH_TO_TALK
                            if (sqwait <= 0 && push) {
                               fprintf(stderr, "\rTalk:  ");
                            }
#endif
                            sqwait = sqdelay;
                            break;
                        }
                    }
                }

                if (squelched) {
                    if (debugging) {
                        printf("Entire buffer squelched.\n");
                    }
                    spurt = TRUE;
                } else {
                    netbuf.compression = fProtocol | (ring ? (fSetDest | fDestSpkr) : 0);
                    netbuf.compression |= debugging ? fDebug : 0;
                    netbuf.compression |= loopback ? fLoopBack : 0;

                    /* If automatic gain control is enabled,
                       ride the gain pot to fill the dynamic range
                       optimally. */

                    if (agc) {
                        register unsigned char *start = bs;
                        register int j;
                        long msamp = 0;

                        for (j = 0; j < soundel; j++) {
                            int s = audio_u2s(*start++);

                            msamp += (s < 0) ? -s : s;
                        }
                        msamp /= soundel;
                        if (msamp < 6000) {
                            if (rgain < 100) {
                                soundrecgain(++rgain);
                            }
                        } else if (msamp > 7000) {
                            if (rgain > 1) {
                                soundrecgain(--rgain);
                            }
                        }
                    }

                    ring = FALSE;
                    if (compressing) {
                        int is = soundel, os = soundel / 2;

                        rate_flow((unsigned char *) buf, (unsigned char *) buf, &is, &os);
                        soundel = os;
                        netbuf.compression |= fComp2X;
                    }
                    netbuf.buffer.buffer_len = soundel;
                    if (!sendpkt(&netbuf)) {
                        exiting();
                    }
                    if (debugging && !vat && !rtp) {
                        fprintf(stderr, "Sent %d audio samples in %ld bytes.\r\n",
                                osl, netbuf.buffer.buffer_len);
                    }
                }
            } else {
                sf_usleep(100000L);  /* Wait for some sound to arrive */
            }
        }
    }
    return 0;
}

/*  MAKESESSIONKEY  --  Generate session key with optional start
                        key.  If mode is TRUE, the key will be
                        translated to a string, otherwise it is
                        returned as 16 binary bytes.  */

static void makeSessionKey(char *key, char *seed, int mode)
{
    int j, k;
    struct MD5Context md5c;
    unsigned char md5key[16], md5key1[16];
#define SESSIONKEYBUF 1024
    char s[SESSIONKEYBUF];
#ifdef HAVE_DEV_RANDOM
    int random_fd, wanted_random_bytes, random_bytes;
    char rand[SESSIONKEYBUF];
#endif

    s[0] = 0;
    if (seed != NULL) {
        strcat(s, seed);
    }

    /* The following creates a seed for the session key generator
       based on a collection of volatile and environment-specific
       information unlikely to be vulnerable (as a whole) to an
       exhaustive search attack.  If one of these items isn't
       available on your machine, replace it with something
       equivalent or, if you like, just delete it. */

    sprintf(s + strlen(s), "%lu", (unsigned long) getpid());
    sprintf(s + strlen(s), "%lu", (unsigned long) getppid());
    V getcwd(s + strlen(s), 256);
    sprintf(s + strlen(s), "%lu", (unsigned long) clock());
    sprintf(s + strlen(s), "%lu", (unsigned long) time(NULL));
#ifdef Solaris
    sysinfo(SI_HW_SERIAL,s + strlen(s), 12);
#else
    sprintf(s + strlen(s), "%lu", (unsigned long) gethostid());
#endif
    getdomainname(s + strlen(s), 256);
    gethostname(s + strlen(s), 256);
    sprintf(s + strlen(s), "%u", getuid());
    sprintf(s + strlen(s), "%u", getgid());

#ifdef HAVE_DEV_RANDOM
    /* Some systems a random device which offers data as long as the system
       has still enough "randomness" collected.  We use it here. */
 
    /* Leave room for the terminator and to append data for second hash. */
    wanted_random_bytes = SESSIONKEYBUF - (strlen(s) - 12);

    if (wanted_random_bytes > 1) {
        dbg("opening /dev/random\n");
        random_fd = open ("/dev/random",O_RDONLY,O_NONBLOCK);
        if (random_fd > 0) { /* Do not complain if open fails */
            random_bytes = read(random_fd, rand, wanted_random_bytes);

            if  (random_bytes  > 0) { /* Again do not complain on read errors */

                if (random_bytes <= wanted_random_bytes) { /* Should be */
                    /* String termination will be lost after memcpy */
                    char *startcopy =  s + strlen(s); 
                    memcpy(startcopy, rand, random_bytes);
                    startcopy[random_bytes + 1] = 0; /* Terminate string */
                }
            }
            close(random_fd);
        }
        dbg("closed /dev/random\n");
    }
#endif

    MD5Init(&md5c);
    MD5Update(&md5c, s, strlen(s));
    MD5Final(md5key, &md5c);
    /*  If you append additional information here, verify that the
        computation of wanted_random_bytes above leaves adequate room after
        the /dev/random data for its maximum length.  */
    sprintf(s + strlen(s), "%lu", (unsigned long) ((time(NULL) + 65121) ^ 0x375F));
    MD5Init(&md5c);
    MD5Update(&md5c, s, strlen(s));
    MD5Final(md5key1, &md5c);
#define nextrand    (md5key[j] ^ md5key1[j])
    if (mode) {
        for (j = k = 0; j < 16; j++) {
            unsigned char rb = nextrand;

/* #define Rad16(x) ((x) + 'A') */
    /* We don't count on choosing where in the alphabet to
       start based on the hashes to stir in any more entropy
       than the fundamental 4 bits from the generated key
       (although it may), but primarily to spread out the
       letter distribution across the entire alphabet rather
       than just the first 16 letters. */
#define Rad16(x)  ((((((md5key1[j] >> 1) ^ (md5key[15 - j] >> 2)) % 26) + (x)) % 26) + 'A')
            key[k++] = Rad16((rb >> 4) & 0xF);
            key[k++] = Rad16(rb & 0xF);
#undef Rad16
            if (j & 1) {
                key[k++] = '-';
            }
        }
        key[--k] = 0;
    } else {
        for (j = 0; j < 16; j++) {
            key[j] = nextrand;
        }
    }
#undef nextrand
}


/*  Main program.  */


int mike_set(int compress, int flags, char *nat, int arobust, int asquelch, int asqdelay){

    adpcmcompress = celpcompress = gsmcompress = lpccompress = lpc10compress = FALSE;
    if (compress & fCompLPC) celpcompress = TRUE;
    if (compress & fComp2X) compressing = TRUE;
    if (compress & fCompADPCM) adpcmcompress = TRUE;
    if (compress & fCompLPC) lpccompress = TRUE;
    if (compress & fCompLPC10) lpc10compress = TRUE; /* TODO robust = ? */
    if (compress & fCompGSM) gsmcompress = TRUE;
    
    if (flags & FLG_DEBUG) debugging = TRUE;
    if (flags & FLG_AGC) agc = TRUE;
    if (flags & FLG_LOOPBACK) loopback = TRUE;
    if (flags & FLG_RING) ring = TRUE;
    if (flags & FLG_RTP) rtp = TRUE;
    if (flags & FLG_VAT) {
        vat = TRUE;
        if (vatid == NULL) {
            vatidl = makeVATid(&vatid, 0L);
        }
    }
    robust = arobust;
    squelch = asquelch;
    sqdelay = asqdelay;

#if 0  
    /* TODO */
    case 'P':             /* -Phost  --  Copy output to host  */
        if (!addest(op + 1)) {
            return 1;
        }
        break;
#endif  
#if 0
    audioDumpFile = fopen(op + 1, "w");
    if (audioDumpFile == NULL) {
        perror("Cannot create raw audio dump (-w option) file");
        return 2;
    }
    break;

#endif
#if 0    
    case 'Y':             /* -Yindev:[ctldev] -- Specify audio input and control device file names
                                                 or #open_fd. */
        devAudioInput = op + 1;
        if (strchr(op + 1, ':') != NULL) {
            devAudioControl = strchr(op + 1, ':') + 1;
        }
        break;
#endif  
#ifdef NAT_LAUNCH
    if (nat){
                                  /* dsock,csock  --  Import file descriptors for sockets */
        char *delim = strchr(nat, ',');
        
        if (delim == NULL) {
            fprintf(stderr, "Error: -NATdata,control option missing control socket FD.\n");
            return 2;
        }
        xnat_socketsharing = TRUE;
        xnat_datasock = atoi(nat);
        xnat_consock = atoi(delim + 1);
        if (debugging) {
            fprintf(stderr,"Using -NAT file descriptors.  Data: %d  Control: %d\n",
                            xnat_datasock, xnat_consock);
        }
    }
#endif

   	return 0; 
}
int mike_start()
{
#ifdef HAVE_LPC    
    static lpcstate_t lpcc;
#endif    
    char md5key[16];
	int i;

	if (mike_pid) {
		dbg("mike already running, stopping\n");
		mike_stop();
	}
	
	if (pipe(mike_pipe)){
		error("Can't create pipe\n");
		return -1;
	}
	
	mike_pid=fork();
	if (mike_pid<0){  /* error */
		error("Can't fork mike\n");
		return -1;
	}
	if (mike_pid>0) { /* parent */
		close(mike_pipe[0]);
		return 0;
	}
	/* child */
	close(mike_pipe[1]);
	close(0);
	if (dup(mike_pipe[0])){
		error("Can't dup\n");
		return -1;
	}
	close(mike_pipe[0]);
	for (i=3;i<16;i++) close(i);

    dbg("mike started (server=%d)\n", opt_s);
	
	
    if (gethostname(hostname, sizeof hostname) == -1) {
        strcpy(hostname, "localhost");
    } else {
        if (strlen(hostname) > 15) {
            hostname[15] = 0;
        }
    }
    
#ifdef OVERDRIVE_SOUND_FILE
    fprintf(stderr, "Warning: iprig compiled with OVERDRIVE_SOUND_FILE set to nonstandard value %g.\n",
        OVERDRIVE_SOUND_FILE);
#endif

#ifdef LINUX_FPU_FIX
    __setfpucw(_FPU_IEEE);            /* Mask floating point interrupts to
                                         enable standard IEEE processing.
                                         Current libc releases do this by
                                         default, so this is needed only on
                                         older libraries. */
#endif
    rate_start(SAMPLE_RATE, SAMPLE_RATE / 2);
    gsmh = gsm_create();

#ifdef ENABLE_LPC     
    
    /* Initialise LPC encoding. */

    if (!lpc_start()) {
        fprintf(stderr, "Cannot allocate LPC decoding memory.\n");
        return 1;
    }
    lpc_init(&lpcc);
    lpc10init();
        
    /* Initialise CELP encoding. */
        
    celp_init(FALSE);
#endif    

#ifdef sgi
    usticks = 1000000 / CLK_TCK;
#endif

#ifdef HALF_DUPLEX
    {
#ifndef INADDR_LOOPBACK
        /* The standard local host loopback address is supposed to be
           defined in <netinet/in.h>, but just in case some screwball
           system doesn't, let's include our own definition here. */
#define INADDR_LOOPBACK  0x7F000001
#endif
        localhost.s_addr = htonl(INADDR_LOOPBACK);
#ifdef HDX_DEBUG
        fprintf(stderr, "iprig: local host %s\n",
                inet_ntoa(localhost));
#endif
    }
#endif

    /* Create the socket used to send data. */

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        perror("opening datagram socket");
        return 1;
    }

    /* Initialise randomised packet counters. */

    makeSessionKey(md5key, NULL, FALSE);
    bcopy(md5key, (char *) &ssrc, sizeof ssrc);
    bcopy(md5key + sizeof ssrc, (char *) &timestamp,
          sizeof timestamp);
    bcopy(md5key + sizeof ssrc + sizeof timestamp,
          (char *) &seq, sizeof seq);
    bcopy(md5key + sizeof ssrc + sizeof timestamp +
          sizeof seq, &rtpdesrand, sizeof rtpdesrand);
    sdesl = rtp_make_sdes(&sdes, ssrc, -1, TRUE);


    if (opt_s){
		/* server */
		struct conn *conn;
		
        foreach(conn, conns){
			addest(conn);
		}
		
	//	addest("192.168.1.97");
		
	}else{
		/* client */
		if (!sconn) error("client: sconn is NULL!\n");
    	addest(sconn);
      //  addest("192.168.1.2");
	}
	
	




    if (rtp && vat) {
        fprintf(stderr, "Cannot choose both -RTP and -VAT protocols simultaneously.\n");
        return 2;
    }

    if (rtp || vat) {
        if (compressing) {
            fprintf(stderr, "Simple (-C) compression cannot be used with -RTP or -VAT protocols.\n");
            return 2;
        }
        if (lpc10compress) {
            fprintf(stderr, "LPC-10 (-LPC10) compression cannot be used with -RTP or -VAT protocols.\n");
            return 2;
        }
        if (celpcompress) {
            fprintf(stderr, "CELP (-CELP) compression cannot be used with -RTP or -VAT protocols.\n");
            return 2;
        }
        if (robust > 1) {
            fprintf(stderr, "Robust (-ROBUSTn) transmission cannot be used with -RTP or -VAT protocols.\n");
            return 2;
        }
        if (ring) {
            fprintf(stderr, "Warning: ring not implemented in -RTP and -VAT protocols.\n");
        }
    }

    return sendfile(NULL);

    exiting();
    gsm_destroy(gsmh);
    lpc_end();
    return 0;
}


int mike_stop(void){
    if (!mike_pid) return 0;

	write(mike_pipe[1], "q", 1);

	dbg("waiting for mike pid %d\n", mike_pid);
	waitpid(mike_pid, NULL, 0);
	dbg("mike pid %d terminated\n", mike_pid);
	mike_pid = 0;
	
	return(0);
}
