/*
    iprig - Remote radio via IP
    Copyright (C) 2007  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct sdev *sdev;

struct sdev *init_sdev(void){
	sdev=mem_alloc(sizeof(struct sdev));
	bzero(sdev, sizeof(struct sdev));
	sdev->fd=-1;
	sdev->rdbuflen=0;
	sdev->conn=NULL;

	return sdev;
}

int sdev_open(void){
	char errbuf[1030];
	struct termios tio;
	
	sdev->rdbuflen=0;
	sdev->conn=NULL;

	sdev->fd=open(SDEV_FILENAME, O_RDWR | O_SYNC | O_NONBLOCK | O_NOCTTY);
	if (sdev->fd < 0){
		error("error opening file %s - %s\n", SDEV_FILENAME, STRERR(errno));
		return -1;
	}
	dbg("sdev opened, fd=%d\n", sdev->fd);
    
	if (fcntl(sdev->fd, F_SETFL, O_NONBLOCK)) {
		error("error setting nonblock %s\n", STRERR(errno));
		sdev_close();
        return -1;
    }
    
	if (tcgetattr(sdev->fd,&tio)){
		error("tcgetattr failed %s\n", STRERR(errno));
		sdev_close();
        return -1;
	}
	tio.c_cflag=B9600|CS8|CLOCAL|CREAD|PARENB;/*|PARODD;*/
	tio.c_iflag=INPCK;
	tio.c_lflag=0;
	tio.c_oflag=0;
	tio.c_cc[VMIN]=1;
	tio.c_cc[VTIME]=5;
    
	if (tcsetattr(sdev->fd,TCSANOW,&tio)){
		error("tcsetattr failed %s\n", STRERR(errno));
		sdev_close();
        return -1;
	}

	return sdev->fd;
}

int sdev_close(void){
	struct termios tio;

	if (sdev->fd<0) return 0;
	sdev->rdbuflen=0;
	sdev->conn=NULL;

	if (tcgetattr(sdev->fd,&tio)!=0){
        tio.c_cflag=B115200|CS8|CLOCAL|CREAD;
        tio.c_iflag=INPCK;
        tio.c_lflag=0;
        tio.c_oflag=0;
        tio.c_cc[VMIN]=1;
        tio.c_cc[VTIME]=5;
        
        tcsetattr(sdev->fd,TCSANOW,&tio);
	}
    
	close(sdev->fd);
	sdev->fd=-1;
	dbg("sdev %d closed\n", sdev->fd);
	return 0;
}

char sd_chk(char *s, int len){
	char chk, *c;
    chk=0;
	for (c=s;len;c++,len--){
		chk^=*c;
	}
	return chk;
}
