/* #define PGP_DEBUG */
/*

                        Speak Freely for Unix
                         Sound Output Server

        Designed and implemented in July of 1991 by John Walker.

*/

#include "speakfree.h"
#include "version.h"
#include "header.h"

static int audiok = FALSE;            /* Audio initialised flag */
static int audiotime = 0;             /* Audio timeout counter */
static int debugging = FALSE;         /* Debugging enabled */
static int debugforce = FALSE;        /* Debugging forced / prevented ? */
#ifdef HEXDUMP
static int hexdump = FALSE;           /* Dump received packets in hex ? */
#endif
static int noring = FALSE;            /* Disable remote ring requests */
gsm gsmh;                             /* GSM handle */
static int whichport = Internet_Port; /* Port to listen on (base address) */
static int sock;                      /* Data socket */
static int ssock;                     /* Control socket (RTP/VAT) */
static struct sockaddr_in from;       /* Sending host address */
static struct sockaddr_in name;       /* Address of destination host */
static socklen_t fromlen;                   /* Length of sending host address */
static int rll;                       /* Length of packet read from socket */
static int showhosts = FALSE;         /* Show host names that connect */
static int hosttimeout = 180 * 1000000L; /* Consider host idle after this time */
static int jitter = 0;                /* Jitter compensation delay */
static int jitterwait = 0;            /* Jitter wait in progress */
static int jitteridle = 0;            /* Countdown to reset jitter timer */
static int jitteridlet = 0;           /* Jitter idle timeout interval */
struct jitterbuf {                    /* Jitter queue item */
    struct jitterbuf *next;
    LONG length;
    char value[2];
};
static struct jitterbuf *jithead = NULL, *jittail = NULL; /* Jitter queue links */
static int outRate = SAMPLE_RATE;     /* Output data rate in samples per second */
static int currentRate = SAMPLE_RATE; /* Current output data rate (possibly dynamically adjusted) */
static FILE *record = NULL;           /* Answering machine record file */
static char *replyfile = NULL;        /* Reply file pathname */
static char *replycmdexe = NULL;
static char *replycmd = "sfmike -t";  /* Default reply command */
static char *busysignal = "sleep 10; sfmike %s busy.au"; /* Default busy signal command */
static int busytimeout = 60 * 1000000L; /* Busy signal timeout interval */
static int dobusy = FALSE;            /* Reject calls when busy ? */
#ifdef HALF_DUPLEX
static struct in_addr localhost;      /* Our internet address */
static int halfDuplexMuted = FALSE;   /* Muted by half-duplex transmission */
#endif

//static struct sockaddr_in lookhost;   /* Look who's listening host, if any */
//static char *sdes = NULL;             /* RTP SDES packet */
//static int sdesl;                     /* RTP SDES packet length */
//static unsigned long ssrc;            /* RTP synchronisation source identifier */
static int lwltimer;                  /* Seconds before next LWL retransmit */
static int lwlonly = FALSE;           /* LWL broadcast only ? */
static int actives = 0;               /* Currently active hosts */

#ifdef NAT_LAUNCH

/* NAT socket-sharing related stuff */

static int xnat_passive = FALSE;      /* Operating in NAT passive mode ? */
static int xnat_active_sfmike_pid = 0;/* Process ID of sfmike we launched */

struct mikeopts_t {
    char *crypto;
    char *compression;
    char *squelch;
    int autogain;
};
#define MIKEOPTS_EMPTY {NULL,NULL,NULL,0}

struct xn_conspec_t {
    char* xnat_conaddr;
    struct xn_conspec_t *next;
};

static struct xn_conspec_t *xnat_conspec = NULL;
#endif

#define LWL_RETRANSMIT  (5 * 60 * 1000000L) /* Microseconds between LWL updates */
#define LWL_STRIKEOUT   5             /* Abandon LWL transmission to site
                                         after this number of consecutive
                                         failures. */

//static struct in_addr lwl_sites[LWL_MAX_SITES]; /* LWL site addresses */
//static long lwl_ports[LWL_MAX_SITES]; /* Ports for LWL hosts */
//static int lwl_strikes[LWL_MAX_SITES];/* Consecutive failures to send LWL packet */
//static int lwl_nsites = 0;            /* Number of LWL sites published on */


static struct connection *conn = NULL;/* Chain of current connections */

static int adjustRate = FALSE;        /* Dynamically adjust output rate ? */
static double timebase;               /* Full seconds at program start time */
static double lastPacketTime = 0;     /* Time last packet arrived */
static double averageDataRate = SAMPLE_RATE; /* Average data arrival rate, samples/second */
static long lastPacketLength = 0;     /* Samples received since last packet */
#define DYNRATE_UPDATE_SAMPLE_COUNT SAMPLE_RATE /* Update dynamic rate after this many samples received */
#define DYNRATE_SMOOTHING_FACTOR 0.1  /* Exponential moving average smoothing factor for rate */

/* Audio output and control device file names. */

#ifdef AUDIO_DEVICE_FILE
extern char *devAudioOutput, *devAudioControl;
#endif

#define Debug   ((debugforce != 0) ? (debugforce > 0) : debugging)

#define TickTock    (10 * 1000000L)   /* Alarm interval in microseconds */
#define TockTock    (60 * 1000000L)   /* Alarm interval when no connections open */
#define JitterTock  (1000000L / 4)    /* Alarm interval for jitter compensation */

#ifdef AUDIO_RELEASE
#define AudioRelease (AUDIO_RELEASE * 1000000L) /* Release audio when idle this long */
#else
#define AudioRelease (20 * 1000000L)  /* Release audio when idle this long */
#endif

static int timerStep = TockTock;      /* Current timer step */

static int speaker_pid = 0;

#define ucase(x)    (islower(x) ? toupper(x) : (x))

#ifdef NAT_LAUNCH

/*  NEW_XNAT_CONSPEC -- Create a new connection specification for
                        later use in launch_sfmike and store it on the list.
*/
#if 0
static struct xn_conspec_t *new_xnat_conspec(char *addr)
{
    struct xn_conspec_t *cur, *neue;
    
    neue = (struct xn_conspec_t *) malloc(sizeof(struct xn_conspec_t));
    if (neue == NULL) {
        fprintf(stderr, "speaker: Unable to allocate NAT connection specification.\n");
        exit(2);
    }
    neue->next = NULL;
    if (xnat_conspec) {
        cur = xnat_conspec;
        while (cur->next != NULL) {
            cur = cur->next;
        }
        cur->next = neue;
    } else {
        xnat_conspec = neue;
    }
    return neue;
}
#endif

static void childterm();

#endif

/*  GSMDECOMP  --  Uncompress the contents of a sound buffer using GSM.  */

static void gsmdecomp(struct soundbuf *sb)
{
    gsm_signal dst[160];
    int i, j, l = 0;
    char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
    static char dcb[BUFL];
    short declen = ntohs(*((short *) sb->buffer.buffer_val));

    /* Since some old releases forget to convert this field
       from network byte order, kludge it back to sanity
       if this happens.  First we try swapping the bytes.  If
       that doesn't work, force the standard buffer length.
       Note that we have to write this out, as ntohs will be
       a no-op on a big-endian machine. */

    if (declen <= 0 || declen > 1600) {
        declen = ((declen & 0xFF) << 8) | ((declen >> 8) & 0xFF);
    }
    if (declen <= 0 || declen > 1600) {
        declen = 1600;
    }
    for (i = 0; i < sb->buffer.buffer_len - sizeof(short);
                i += sizeof(gsm_frame)) {
        gsm_decode(gsmh, (gsm_byte *) dpx, dst);
        dpx += sizeof(gsm_frame);
        for (j = 0; j < 160; j++) {
            dcb[l++] = audio_s2u(dst[j]);
        }
    }
    bcopy(dcb, sb->buffer.buffer_val, declen);
    sb->buffer.buffer_len = declen;
}

#ifdef ENABLE_LPC
/*  LPCDECOMP  --  Uncompress the contents of a sound buffer using LPC.  */

static void lpcdecomp(struct soundbuf *sb, struct connection *c)
{
    int i, l = 0;
    char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
    char dcb[BUFL];
    short declen = ntohs(*((short *) sb->buffer.buffer_val));

    if (declen <= 0 || declen > (10 * LPC_FRAME_SIZE)) {
        declen = 10 * LPC_FRAME_SIZE;
    }
    for (i = 0; l < declen;
                i += LPCRECSIZE) {
        lpc_synthesize(dcb + l, (lpcparams_t *) (dpx + i), &c->lpcc);
        l += LPC_FRAME_SIZE;
    }
    bcopy(dcb, sb->buffer.buffer_val, declen);
    sb->buffer.buffer_len = declen;
}

/*  LPC10DECOMP  --  Uncompress the contents of a sound buffer using LPC10.  */

static void lpc10decomp(struct soundbuf *sb, struct connection *c)
{
    int j;
    unsigned char *dpx = ((unsigned char *) sb->buffer.buffer_val);
    unsigned char dcb[BUFL];

    j = lpc10decode(dpx, dcb, sb->buffer.buffer_len);
    bcopy(dcb, sb->buffer.buffer_val, j);
    sb->buffer.buffer_len = j;
}

/*  CELPDECOMP  --  Uncompress the contents of a sound buffer encoded in CELP.  */

static void celpdecomp(struct soundbuf *sb)
{
    short dst[240];
    int i, j, l = 0;
    char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
    static char dcb[BUFL];
    short declen = ntohs(*((short *) sb->buffer.buffer_val));

    if (declen <= 0 || declen > 1920) {
        declen = 1920;
    }
    for (i = 0; i < sb->buffer.buffer_len - sizeof(short);
                i += 18) {
        celp_decode(dpx, dst);
        dpx += 18;
        for (j = 0; j < 240; j++) {
            dcb[l++] = audio_s2u(dst[j]);
        }
    }
    bcopy(dcb, sb->buffer.buffer_val, declen);
    sb->buffer.buffer_len = declen;
}
#endif

/*  ETIME  --  Edit time and date for log messages.  */

static char *etime(void)
{
    struct tm *t;
    time_t clock;
    static char s[20];

    time(&clock);
    t = localtime(&clock);
    sprintf(s, "%02d-%02d %02d:%02d", t->tm_mon + 1, t->tm_mday,
               t->tm_hour, t->tm_min);
    return s;
}

/*  COMPRESSIONTYPE  --  Return a string describing the type of
                         compression employed in this buffer.  Assumes
                         this *is*, in fact, a sound buffer and not
                         an RTP packet, face data, etc.  */

static char *compressionType(soundbuf *msg)
{
    int compression;

    compression = ntohl(msg->compression);
    return ((compression & (fComp2X | fCompGSM)) ==
                           (fComp2X | fCompGSM)) ?
                           "GSM+2X compressed" :
           ((compression & (fComp2X | fCompADPCM)) ==
                           (fComp2X | fCompADPCM)) ?
                           "ADPCM+2X compressed" :
           ((compression & (fComp2X | fCompLPC)) ==
                           (fComp2X | fCompLPC)) ?
                           "LPC+2X compressed" :
           ((compression & (fComp2X | fCompLPC10)) ==
                           (fComp2X | fCompLPC10)) ?
                           "LPC10+2X compressed" :
           ((compression & (fComp2X | fCompCELP)) ==
                           (fComp2X | fCompCELP)) ?
                           "CELP+2X compressed" :
           ((compression & (fComp2X | fCompVOX)) ==
                           (fComp2X | fCompVOX)) ?
                           "VOX+2X compressed" :
           ((compression & fCompADPCM) ? "ADPCM compressed" :
           ((compression & fCompLPC) ? "LPC compressed" :
           ((compression & fCompLPC10) ? "LPC10 compressed" :
           ((compression & fCompCELP) ? "CELP compressed" :
           ((compression & fComp2X) ? "2X compressed" :
           ((compression & fCompGSM) ? "GSM compressed" :
           ((compression & fCompVOX) ? "VOX compressed" :
                                       "uncompressed")))))));
}

#if 0
/*  MAKESESSIONKEY  --  Generate session key.  */

static void makeSessionKey(char *key)
{
    int j;
    struct MD5Context md5c;
    unsigned char md5key[16], md5key1[16];
    char s[1024];

    s[0] = 0;
    sprintf(s + strlen(s), "%u", getpid());
    sprintf(s + strlen(s), "%u", getppid());
    V getcwd(s + strlen(s), 256);
    sprintf(s + strlen(s), "%lu", (unsigned long) clock());
    sprintf(s + strlen(s), "%lu", (unsigned long) time(NULL));
#ifdef Solaris
    sysinfo(SI_HW_SERIAL, s + strlen(s), 12);
#else
    sprintf(s + strlen(s), "%lu", (unsigned long) gethostid());
#endif
    getdomainname(s + strlen(s), 256);
    gethostname(s + strlen(s), 256);
    sprintf(s + strlen(s), "%u", getuid());
    sprintf(s + strlen(s), "%u", getgid());
    MD5Init(&md5c);
    MD5Update(&md5c, s, strlen(s));
    MD5Final(md5key, &md5c);
    sprintf(s + strlen(s), "%lu", (time(NULL) + 65121) ^ 0x375F);
    sprintf(s + strlen(s), "%lu", (unsigned long) (getppid() ^ clock()));
    MD5Init(&md5c);
    MD5Update(&md5c, s, strlen(s));
    MD5Final(md5key1, &md5c);
    for (j = 0; j < 16; j++) {
        key[j] = md5key[j] ^ md5key1[j];
    }
}
#endif

/*  ULARM  --  Wrapper for setitimer() that looks like alarm()
               but accepts a time in microseconds.  */

static void ularm(long t)
{
    struct itimerval it;

    it.it_value.tv_sec = t / 1000000L;
    it.it_value.tv_usec = t % 1000000L;
    it.it_interval.tv_sec = it.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &it, NULL);
}

/*  WINDTIMER  --  Reset alarm timer to appropriate value
                   depending on what's going on at the moment.  */

static void windtimer(void)
{
    if (audiok || (actives > 0)) {
        timerStep = (jitteridle > 0) ? JitterTock : TickTock;
    } else {
        timerStep = TockTock;
    }
    ularm(timerStep);
}

/*  ISBUSY  --  Test if a connection is active (item on connection
                chain with a protocol other than PROTOCOL_UNKNOWN).  */

static int isBusy(void)
{
    struct connection *c;

    c = conn;
    while (c != NULL) {
        if (c->con_protocol != PROTOCOL_UNKNOWN) {
            return dobusy;
        }
        c = c->con_next;
    }
    return FALSE;
}

/*  SENDLWLMESSAGE  --  If enabled, send a message identifying us
                        to each selected Look Who's Listening server.  */

#if 0
static void sendLwlMessage(int dobye)
{
    int i, sock;

    for (i = 0; i < lwl_nsites; i++) {
        if (lwl_ports[i] >= 0) {
            if (Debug) {
                fprintf(stderr, "speaker: updating LWL data at %s\n",
                    inet_ntoa(lwl_sites[i]));
            }
            sock = socket(AF_INET, SOCK_STREAM, 0);
            if (sock < 0) {
                perror("opening look who's listening socket");
                //sdes = NULL;
                return;
            }

            lookhost.sin_port = htons(lwl_ports[i]);
            bcopy((char *) (&lwl_sites[i]), (char *) &lookhost.sin_addr.s_addr,
                  sizeof lookhost.sin_addr.s_addr);

            if (connect(sock, (struct sockaddr *) &(lookhost), sizeof lookhost) >= 0) {
                if (dobye) {
                    unsigned char v[1024];
                    int l;

                    l = rtp_make_bye(v, ssrc, "Exiting sfspeaker", FALSE);
                    if (send(sock, v, l, 0) < 0) {
                        perror("sending look who's listening BYE packet");
                    }
                    /* No point in clearing strikes if we're exiting. */
                } else {
                    if (send(sock, (char *) sdes, sdesl, 0) < 0) {
                        perror("sending look who's listening source ID message");
                    } else {
                        lwl_strikes[i] = 0; /* Success--clear strikes */
                    }
                }
            } else {
                perror("connecting look who's listening socket");

                /* If we get a connection refused reply, increment the number
                   of strikes against the server.  If the server strikes out,
                   remove it from the list of server to which we send our
                   LWL information. */

                if (errno == ECONNREFUSED) {
                    if ((++lwl_strikes[i]) >= LWL_STRIKEOUT) {
                        if (Debug) {
                            fprintf(stderr, "speaker: abandoning LWL transmission to %s\n",
                                inet_ntoa(lwl_sites[i]));
                        }
                        lwl_ports[i] = -1;
                    }
                }
            }
            close(sock);
        }
    }
    lwltimer = LWL_RETRANSMIT;
}
#endif

/*  OBTAINAUDIO  --  Attempt to obtain the audio output device.  */

static void obtainaudio(void)
{
    if (!audiok) {
        if (!soundinit(O_WRONLY)) {
            perror("opening audio output device");
            fprintf(stderr, "A common cause of this error is a sound board or\n");
            fprintf(stderr, "driver which cannot run in full duplex mode (some\n");
            fprintf(stderr, "boards which are physically capable of full duplex\n");
            fprintf(stderr, "have drivers which cannot run them in this mode).\n\n");
            fprintf(stderr, "Try uncommenting the line:\n\n");
            fprintf(stderr, "    DUPLEX = -DHALF_DUPLEX\n\n");
            fprintf(stderr, "in the Makefile, then \"make clean\", then \"make\".\n");
            fprintf(stderr, "This will operate your board in half-duplex mode\n");
            fprintf(stderr, "and avoid the conflict which probably caused the\n");
            fprintf(stderr, "error opening audio output.\n");
            return;                   /* Can't acquire sound device */
        }
        audiok = TRUE;
        if (Debug) {
            fprintf(stderr, "speaker: opening audio device.\n");
        }
        windtimer();                  /* Reset timer, if necessary */
    }
}

/*  FLUSHJITTER  --  Flush any buffers waiting in the jitter compensation
                     queue.  */

static void flushjitter(void)
{
    struct jitterbuf *jb;

    /* If any samples are pending in the jitter queue, flush
       them to the audio output.  */

    if ((jb = jithead) != NULL) {
        jithead = jittail = NULL;

        if (Debug) {
            fprintf(stderr, "Flushing jitter queue.\n");
        }
        obtainaudio();
        while (jb != NULL) {
            struct jitterbuf *jo = jb;

            if (Debug) {
                fprintf(stderr,
                        "Playing %ld samples from jitter queue.\n", jb->length);
            }
            if (audiok) {
                audiotime = 0;
                soundplay(jb->length, (unsigned char *) jb->value);
            }
            jb = jb->next;
            free(jo);
        }
    }
}

/*  RELEASE  --  Signal-catching function which releases the audio
                 device if we haven't received anything to play in
                 the last minute. */

static void release()
{
    struct connection *c, *l, *n;

    /* Mark idle any hosts that haven't sent us anything recently. */

    c = conn;
    l = NULL;
    actives = 0;
    while (c != NULL) {
        if ((c->con_timeout >= 0) && (c->con_timeout < hosttimeout)) {
            c->con_timeout += timerStep;
            actives++;
        }

        /* If the busy signal timer is running, decrement it.
           Once it expires the host becomes eligible to attempt
           to connect once again.  In essence, the connection
           is held open until the busy signal timeout expires, and
           is then forced to go idle with the regular timeout
           mechanism, making it eligible once again to attempt
           connection. */

        if (c->con_busy > 0) {
            c->con_busy -= timerStep;
            if (c->con_busy <= 0) {
                c->con_busy = 0;
                c->con_timeout = hosttimeout;
                if (Debug) {
                    fprintf(stderr, "speaker: %s %s busy signal timeout expired\n", etime(), c->con_hostname);
                }
            }
        }
        n = c->con_next;
        if (c->con_timeout >= hosttimeout) {
            actives--;                /* Oops--just went inactive */
            c->con_timeout = -1;      /* Mark inactive */
            if (showhosts) {
                fprintf(stderr, "speaker: %s %s idle\n", etime(), c->con_hostname);
            }
            c->con_protocol = PROTOCOL_UNKNOWN;
            c->con_uname[0] = c->con_email[0] = 0;
            c->con_rseq = -1;


            /* If there's a PGP session key for this host, don't actually
               release the connection buffer, as that would lose it.  Issue
               an idle message, though, and tweak the timeout so we do this
               only once. */

            if (c->pgpkey[0] == 0) {
                if (l == NULL) {
                    conn = n;
                } else {
                    l->con_next = n;
                }
                free(c);
            }
        } else {
            l = c;
        }
        c = n;
    }

    /* Release the sound device after two ticks.  This allows
       other programs to use it while we're idle. */

    if (audiok && ((audiotime += timerStep) >= AudioRelease)) {
        soundterm();
        audiok = FALSE;
        if (Debug) {
            fprintf(stderr, "speaker: releasing audio device.\n");
        }
        
        /* When audio goes idle, reset the dynamic data rate computation
           so the rate from the last connection or session doesn't bias
           that of a subsequent connection. */
           
        lastPacketTime = 0;
        averageDataRate = SAMPLE_RATE;
        lastPacketLength = 0;

        /* Flush the record file so if we're rudely terminated it
           will be up to date as of the last audio idle time. */

        if (record != NULL) {
            fflush(record);
        }
    }

    /* Update our Look Who's Listening information if the
       timeout has expired. */

    lwltimer -= timerStep;
/*    if (sdes != NULL && lwltimer <= 0) {
        sendLwlMessage(FALSE);
    }*/

    /* If sound is waiting in the jitter queue, decrement the
       jitter timer and flush the queue if it decrements to zero. */

    if (jithead != NULL && jitterwait > 0) {
        jitterwait -= timerStep;
        if (jitterwait <= 0) {
            jitterwait = 0;
            flushjitter();
        }
    }

    /* Count down the jitter idle timer.  When it's zero, the
       first packets of an audio stream will be buffered in
       the jitter queue. */

    if (jitteridle > 0) {
        jitteridle -= timerStep;
        if (jitteridle <= 0) {
            jitteridle = 0;
            if (Debug) {
                fprintf(stderr, "speaker: jitter idle\n");
            }
        }
    }

    /* If we still own the sound device or have a connection
       open, reset the timer. */

    windtimer();
    if (Debug) {
        time_t t = time(NULL);
        fprintf(stderr, "speaker: Tick: %.2f... %s",
                timerStep / 1000000.0, ctime(&t));
    }
    signal(SIGALRM, release);         /* Set signal to handle timeout */
}

/*  EXITING  --  Catch as many program termination signals as
                 possible and clean up before exit.  */

static void exiting()
{

/*    if (sdes) {
        sendLwlMessage(TRUE);
    }*/

    exit(0);
}

/*  CHILDTERM   --  Handle termination of a face viewer child
                    process and any sfmikes.  This is necessary both to avoid
                    viewers the user terminates manually becoming
                    zombie processes, and also to delete the
                    face file no longer associated with a viewer. */

static void childterm()
{
    int status, pid;

    while ((pid = 
#ifdef Solaris
                  /* Should work for any System V Unix. */
                  waitpid(0 /* or -1 */,&status, WNOHANG)
#else
                  /* BSD and IRIX implement wait3(). */
                  wait3(&status, WNOHANG, 0)
#endif
           ) > 0) {
        if (Debug) {
            fprintf(stderr, "Child process %d done\n", pid);
        }
        if (WIFEXITED(status) || WIFSIGNALED(status)) {
            struct connection *c = conn;

            while (c != NULL) {

#ifdef NAT_LAUNCH
                if (pid == c->sfmike_pid) {
                    if (Debug) {
                        if (WIFEXITED(status))
                            fprintf(stderr,  "sfmike terminated with status %d\n",
                                    WEXITSTATUS(status));
                        if (WIFSIGNALED(status))
                            fprintf(stderr, "sfmike terminated with signal %d\n",
                                    WTERMSIG(status));
                    }
                    c->sfmike_pid = 0;
                    printf("Transmission terminated. Continue listening [y/n]? ");
                    fflush(stdout);
                    if (getc(stdin) == 'n') {
                        exit(1);
                    }
                }
#endif
                c = c->con_next;
            }
        }
    }
    signal(SIGCHLD, childterm);         /* Set signal to handle face viewer termination */
}

/*  PLAYBUFFER  --  Send next buffer to audio output. */

static void playbuffer(soundbuf *msg, struct connection *c)
{
    char *val;
    LONG len;
    unsigned long cmodes;
    char auxbuf[BUFL + 2]/*, bbuf[8], tbuf[8]*/;
    struct timeval tp;
    struct timezone tzp;
    double packetTime = 0;

    if (adjustRate) {
        gettimeofday(&tp, &tzp);          /* Time stamp this packet's arrival time, in microseconds */
        packetTime = (tp.tv_sec - timebase) * 1000000.0 + tp.tv_usec;
/*fprintf(stderr, "Packet = %.0f, lastPacket = %.0f, delta = %.0f\n", packetTime, lastPacketTime, packetTime - lastPacketTime);*/
    }

    debugging = (msg->compression & fDebug) ? TRUE : FALSE;
    
    audiotime = 0;                    /* Reset timeout counter */
    obtainaudio();                    /* Open audio output if necessary */

    if (showhosts && (c->con_compmodes != (fCompressionModes &
            msg->compression))) {
        c->con_compmodes = fCompressionModes & msg->compression;
        fprintf(stderr, "speaker: %s sending %s.\n", c->con_hostname,
                compressionType(msg));
    }

    len = msg->buffer.buffer_len;
    val = msg->buffer.buffer_val;
    
    /*  Long we get here, we should have already verified that the
        length in the packet is consistent with the length of
        data actually read from the socket which, in turn, is
        guaranteed to be limited to BUFL--our maximum buffer size.
        Just as a sanity check, and prior to the specific checks
        made below based on details of the protocol, make sure
        something didn't slip by which might cause us to index
        off the end of a buffer.  We allow for worst-case round-up
        of the size in the buffer due to encryption in making
        the comparison.  */
        
    if (((len + 15) & (~15)) >= BUFL) {
        fprintf(stderr, "speaker: ** Sound buffer from %s in %s of %ld bytes\n"
                        "       exceeds maximum size of %d bytes--discarded. **\n",
            c->con_hostname, compressionType(msg), len, BUFL);
        return;
    }

    if (Debug) {
        fprintf(stderr, "speaker: playing %ld %s bytes from %s.\n", len,
                compressionType(msg), c->con_hostname);
    }

    /* If the fSetDest bit is on, use the fDestJack bit to re-route
       the sound.  This is normally used to divert the sound to the
       speaker to get an individual's attention.  This can be
       disabled with the -N option.  */

    if ((msg->compression & fSetDest) && !noring) {
        sounddest((msg->compression & fDestJack) ? 1 : 0);
        if (!(msg->compression & fDestJack)) {
            if (audiok) {
                soundplayvol(50);     /* Make sure volume high enough */
            }
        }
    }

    if ((msg->compression & (fEncDES | fEncOTP | fEncIDEA | fEncPGP | fEncBF | fEncAES)) &&
        !(c->con_crypt_warning)) {
        c->con_crypt_warning = TRUE;
        fprintf(stderr, "speaker: Warning: host %s is sending encrypted audio\n",
                c->con_hostname);
        fprintf(stderr, "   which cannot be decoded by this version of\n");
        fprintf(stderr, "   Speak Freely, which was built with encryption\n");
        fprintf(stderr, "   removed to permit redistribution without concern\n");
        fprintf(stderr, "   for export control and other regulations.\n");
        fprintf(stderr, "   You can download a version of Speak Freely which\n");
        fprintf(stderr, "   includes full encryption from:\n");
        fprintf(stderr, "       http://www.fourmilab.ch/speakfree/unix/\n");
    }

    /* With the exception of 2X compression, which may be selected
       by itself or combined with any other mode, compression modes
       are mutually exclusive.  Verify that this is the case in the
       packet before proceeding with decompression.  Any possible
       damage which may be caused by this will be averted by the
       length checks in the decoding sequence below, but identifying
       the source of the error before we blither into the
       decoders makes diagnosis easier, which is important since
       packets specifying multiple compression modes are an obvious
       way malicious users may try to provoke buffer overflows. */
       
    cmodes = msg->compression & (fCompressionModes ^ fComp2X);
    if (cmodes != 0) {
        while ((cmodes & 1) == 0) {
            cmodes >>= 1;
        }
        if (cmodes != 1) {
            fprintf(stderr, "speaker: ** Audio packet from %s specifies multiple mutually "
                            "exclusive compression modes: %08lX.  Dropped. **\n",
                            c->con_hostname, msg->compression);
            return;
        }
    }

    /* If message is compressed, decompress appropriately.  Here we get
       quite strict enforcing packet size constraints.  Regardless of the
       provenance of the packet, depending on its compression and protocol
       there is a maximum packet size which will never be exceeded by any
       valid message, and these sizes are well below the threshold at which
       a buffer overflow might occur.  Based on the compression mode, we
       check against the maximum and discard the packet if the contents
       are too long to be valid.  For some protocols we can also test that
       the payload is an even number of encoded frames in the specified
       compression mode.  Notwithstanding these checks, the CODECs still must
       be invulnerable to garbage data which passes these length tests.
       
       Maximum packet payload lengths, by compression type and protocol,
       are as follows.  Protocols with a fixed frame size give the frame
       size after the protocol name.  Many compression types embed the
       actual decompressed length in the first two bytes of the payload,
       accounting for the total being two bytes greater than an even number
       of frames.  ADPCM appends its 3 byte state to the end of the compressed
       audio stream.

                            Speak Freely        RTP         VAT
                            
        Uncompressed            488             320         320
       
        Simple                  488
       
        GSM     [33]            332             134         134
            +Simple             332
         
        ADPCM                   489             323         323
            +Simple             490
        
        LPC     [14]            142              58          58
            +Simple             142
            
        LPC10    [7]             70 
            +Simple              70
            
        CELP    [18]            146
            +Simple             146
            
    */

#define Checklen(n, t)  if (msg->buffer.buffer_len > (n)) { \
            fprintf(stderr, "speaker: ** Payload (%ld) bytes from %s exceeds maximum of %d bytes for " \
                            t " compression. **\n", msg->buffer.buffer_len, c->con_hostname, n); \
            return; }

#define Checkframe(f, p, t)    if (((msg->buffer.buffer_len - (p)) % (f)) != 0) { \
            fprintf(stderr, "speaker: ** Payload (%ld+%d pad) bytes from %s is not an even number of %d byte frames for " \
                            t " compression. **\n", msg->buffer.buffer_len - (p), p, c->con_hostname, f); \
            return; }

    if (msg->compression & fCompVOX) {
        Checklen(332, "VOX/GSM");
        Checkframe(33, 2, "VOX/GSM");
        vox_gsmdecomp(msg);
        len = msg->buffer.buffer_len;
    }

    if (msg->compression & fCompGSM) {
        Checklen(332, "GSM");
        Checkframe(33, 2, "GSM");
        gsmdecomp(msg);
        len = msg->buffer.buffer_len;
    }

    if (msg->compression & fCompADPCM) {
        Checklen(490, "ADPCM");
        adpcmdecomp(msg);
        len = msg->buffer.buffer_len;
    }

#ifdef ENABLE_LPC    
    if (msg->compression & fCompLPC) {
        Checklen(142, "LPC");
        Checkframe(14, 2, "LPC");
        lpcdecomp(msg, c);
        len = msg->buffer.buffer_len;
    }

    if (msg->compression & fCompLPC10) {
        Checklen(70, "LPC10");
        Checkframe(7, 0, "LPC10");
        lpc10decomp(msg, c);
        len = msg->buffer.buffer_len;
    }

    if (msg->compression & fCompCELP) {
        Checklen(146, "CELP");
        Checkframe(18, 2, "CELP");
        celpdecomp(msg);
        len = msg->buffer.buffer_len;
    }
#endif    

    if (msg->compression & fComp2X) {
        int is = len, os = len * 2;

        /*  The following Checklen must be sufficiently large to handle
            the maximum result of any of the above decoders.  At the moment,
            this is the 1920 bytes resulting from CELP+2X compression.  */
        Checklen(1920, "2X");
        rate_start(SAMPLE_RATE / 2, SAMPLE_RATE);
        rate_flow((unsigned char *) val, (unsigned char *) auxbuf, &is, &os);
        bcopy(auxbuf, val, os);
        len = os;
    }
    
    /* Update the dynamic sample arrival rate and the exponentially
       smoothed moving average of the same.  To avoid short-term jitter,
       we first average the arrival rate by computing it only after a
       total of DYNRATE_UPDATE_SAMPLE_COUNT samples have arrived (in
       however many packets it takes).  These individual measures
       are then smoothed with an exponentially smoothed moving
       average with smoothing factor DYNRATE_SMOOTHING_FACTOR to
       determine the average arrival rate. */
       
    lastPacketLength += len;
    if (adjustRate && (lastPacketLength > DYNRATE_UPDATE_SAMPLE_COUNT)) {
        if (lastPacketTime != 0) {
            int oldRate;
            double idrate = lastPacketLength / ((packetTime - lastPacketTime) / 1e6);   /* Instantaneous data rate */
            
            if ((idrate > ((SAMPLE_RATE * 9) / 10)) && (idrate < ((SAMPLE_RATE * 11) / 10))) {
/*
fprintf(stderr, "lastPacketLength = %ld, elapsedTime = %.6f\n", lastPacketLength, (packetTime - lastPacketTime) / 1e6);
*/
                averageDataRate = averageDataRate + ((idrate - averageDataRate) * DYNRATE_SMOOTHING_FACTOR);
                if (Debug) {
                    fprintf(stderr, "speaker: data rate (samples/sec): instantaneous %.0f, average %.0f\n", 
                        idrate, averageDataRate);
                }
                
                /* Now adjust the output rate to dispose of data at the
                   smoothed arrival rate of input.  Note that we only adjust
                   the rate if it is greater than the nominal SAMPLE_RATE; this
                   creates the bias which prevents secular growth in the output
                   queue.  Further, the rate we compute is based on outRate,
                   which can be adjusted by the -fx.x option to compensate for
                   a clock on the audio hardware which has a bias with regard to
                   the CPU clock we use to time sample arrival rate. */
                   
                oldRate = currentRate;

                currentRate = (averageDataRate <= SAMPLE_RATE) ? SAMPLE_RATE :
                                ((outRate * SAMPLE_RATE) / averageDataRate);
                if (Debug && (currentRate != oldRate)) {
                    fprintf(stderr, "speaker: dynamically adjusting output rate to %d samples/sec.\n", currentRate);
                }
            }
/*
else {
fprintf(stderr, "Ignored ridiculous instantaneous data rate of %.0f\n", idrate);
}
*/
        }
        lastPacketTime = packetTime;
        lastPacketLength = 0;
    }
    
    /* If output rate adjustment is requested, re-sample the data
       in the sound buffer to the requested rate.  If the requested
       rate would cause the output buffer to overflow, the length is
       limited to that of the buffer, truncating any additional
       audio in the packet.  In practice, this only happens with
       absurdly large rate reductions which would render the audio
       incomprehensible in any case. */
           
    if (currentRate != SAMPLE_RATE) {
        int is = len, os = ((len * currentRate) + 7999) / SAMPLE_RATE;

        if (os > BUFL) {
            os = BUFL;          /* Constrain length to output buffer size */
        }
        rate_start(SAMPLE_RATE, currentRate);
        rate_flow((unsigned char *) val, (unsigned char *) auxbuf, &is, &os);
        bcopy(auxbuf, val, os);
        len = os;
        if (Debug) {
            fprintf(stderr, "speaker: resampled to %ld samples at %d samples/sec.\n", len, currentRate);
        }       
    }

    /* If the jitter idle time has elapsed, reset the jitter
       wait to queue the start of this packet sequence. */

    if (jitter > 0 && jitteridle == 0 && jitterwait == 0) {
        jitterwait = jitter;
        if (Debug) {
            fprintf(stderr, "speaker: starting %ld millisecond jitter delay.\n", jitterwait / 1000L);
        }
    }

    /* Receipt of any sound buffer resets the jitter idle timer. */

    jitteridle = jitteridlet;         /* Reset jitter idle countdown */

    /* If the jitter wait has not yet expired, add the samples to
       the queue waiting to be played at the end of the jitter
       wait.  */

    if (jitterwait > 0) {
        struct jitterbuf *jb;

        jb = (struct jitterbuf *) malloc(sizeof(struct jitterbuf) + len);
        if (jb != NULL) {
            bcopy(val, jb->value, len);
            jb->length = len;
            jb->next = NULL;
            if (jittail != NULL) {
                jittail->next = jb;
            }
            jittail = jb;
            if (jithead == NULL) {
                jithead = jb;
                windtimer();
            }
            if (Debug) {
                fprintf(stderr, "speaker: adding %ld samples to jitter queue.\n",
                        len);
            }
        }
    } else {
        flushjitter();
        if (audiok) {
            soundplay(len, (unsigned char *) val);
        }
    }
    if (record != NULL) {
        fwrite(val, len, 1, record);
    }
}


/*  Main program.  */

void dump_packet(struct soundbuf *sb){

        int compr;
        compr = sb->compression;
        dbg("compression: %08lx ", compr);
        if (compr & fComp2X) dbg("2X ");
        if (compr & fDebug)         dbg("debug ");
        if (compr & fSetDest)       dbg("setdest ");
        if (compr & fDestSpkr)      dbg("spkr ");
        if (compr & fDestJack)      dbg("jack ");
        if (compr & fLoopBack)      dbg("loopback ");
        if (compr & fCompGSM)       dbg("gsm ");
        if (compr & fEncDES)        dbg("des ");
        if (compr & fEncOTP)        dbg("otp ");
        if (compr & fEncIDEA)       dbg("idea ");
        if (compr & fCompADPCM)     dbg("adpcm ");
        if (compr & fEncPGP)        dbg("pgp ");
        if (compr & fKeyPGP)        dbg("pgpkey ");
        if (compr & fCompLPC)       dbg("lpc ");
        if (compr & fCompVOX)       dbg("vox ");
        if (compr & fCompLPC10)     dbg("lpc10 ");
        if (compr & fCompRobust)    dbg("robust ");
        if (compr & fEncBF)         dbg("bf ");
        if (compr & fCompCELP)      dbg("celp ");
        if (compr & fEncAES)        dbg("aes ");
        if (compr & fProtocol)      dbg("protocol ");
        dbg("\n");
}

int speaker_set(int flags, int port, int ajitter, int ajitteridlet){
    
    if (flags & FLG_DEBUG) debugforce = TRUE;
    if (flags & FLG_HEXDUMP) {
        debugforce = TRUE;
        hexdump = TRUE;
    }
    if (port>0) whichport = port;

    jitter = ajitter;
    jitteridlet = (ajitteridlet == 0) ? (jitter * 2) : ajitteridlet;
    if (jitter < 100000) {
        fprintf(stderr, "speaker: invalid jitter wait %d--reset to 100 milliseconds.\n", jitter / 1000);
        jitter = 100000;
    }
    if (jitteridlet < 100000) {
        fprintf(stderr, "speaker: invalid jitter idle time %d--reset to 100 milliseconds.\n", jitteridlet / 1000);
        jitteridlet = 100000;
    }

#if 0
#ifdef NAT_LAUNCH
                      /* -NAThost[:port]  --  Connect to NAT host */
    if ((ucase(op[1]) == 'A') && (ucase(op[2]) == 'T')) {
        if (strlen(op + 3) > 0) {
            cur_xnat_conspec = new_xnat_conspec(op + 3);
            Gbuf(cur_xnat_conspec->xnat_conaddr, char *, strlen(op + 3) + 1);
            strcpy(cur_xnat_conspec->xnat_conaddr, op + 3);
        } else {
            xnat_passive = TRUE;
        }
                      /* -NCmode  --  Specify compression mode for NAT connection */
    } else if (ucase(op[1]) == 'C') {
        if (strlen(op + 2) > 0) {
            if (mikeopts == NULL) {
                Gbuf(mikeopts, struct mikeopts_t *, sizeof(struct mikeopts_t));
            }
            Gbuf(mikeopts->compression, char *, strlen(op + 2) + 2);
            strcpy(mikeopts->compression, "-");
            strcat(mikeopts->compression, op + 2);
        }
    } else {
#endif
        noring = TRUE;
#ifdef NAT_LAUNCH
    }
#endif
#endif    

#if 0
#ifdef AUDIO_DEVICE_FILE
                case 'Y':             /* -Youtdev:[ctldev] -- Specify audio
                                                     output and control device
                                                     file names or #open_fd. */
                    devAudioOutput = op + 1;
                    if (strchr(op + 1, ':') != NULL) {
                        devAudioControl = strchr(op + 1, ':') + 1;
                    }
                    break;
#endif
#endif    
    return 0;
}


int speaker_start()
{
    size_t length;
    struct soundbuf sb;
    /*FILE *fp;*/
    struct connection *c;
    int newconn, wasrtp;
    int i;

    /* Save the number of full seconds at progam start time.  We subtract this
       from packet arrival times measured in microseconds to minimise floating
       point round-off in subsequent computations. */
    {
        struct timeval tp;
        struct timezone tzp;
        gettimeofday(&tp, &tzp);
        timebase = tp.tv_sec;
    }


	if (speaker_pid) {
		dbg("speaker already running, stopping\n");
		speaker_stop();
	}
        
        /* Create the sockets from which to read */
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        perror("opening data socket");
        return 1;
    }
    
    ssock = socket(AF_INET, SOCK_DGRAM, 0);
    if (ssock < 0) {
        perror("opening control socket");
        return 1;
    }

    /* Create name with wildcards. */

    name.sin_family = AF_INET;
    name.sin_addr.s_addr = INADDR_ANY;
    name.sin_port = htons(whichport);
    if (bind(sock, (struct sockaddr *) &name, sizeof name) < 0) {
        perror("binding data socket");
        return 1;
    }
    name.sin_port = htons(whichport + 1);
    if (bind(ssock, (struct sockaddr *) &name, sizeof name) < 0) {
        perror("binding control socket");
        return 1;
    }
    name.sin_port = htons(whichport);

	
	speaker_pid=fork();
	if (speaker_pid<0){  /* error */
		error("Can't fork speaker\n");
		return -1;
	}
	if (speaker_pid>0) { /* parent */
        close(sock);
        close(ssock);
		return 0;
	}
	for (i=3;i<16;i++) {
        if (i!=sock && i!=ssock) close(i);
    }

    dbg("speaker started\n");
    
    if (!lwlonly) {




        /*  Process command line options.  */


#if Internet_Port != 2074
        /* This is a reminder to the author who frequently tests with
           non-standard ports to avoid inadvertently releasing a
           production version configured for the wrong port. */
        fprintf(stderr, "%s: warning, listening on non-standard Internet port %d\n",
            prog, Internet_Port);
#endif


#ifdef HALF_DUPLEX
        {
#ifndef INADDR_LOOPBACK
        /* The standard local host loopback address is supposed to be
           defined in <netinet/in.h>, but just in case some screwball
           system doesn't, let's include our own definition here. */
#define INADDR_LOOPBACK  0x7F000001
#endif
        localhost.s_addr = htonl(INADDR_LOOPBACK);
#ifdef HDX_DEBUG
            if (Debug) {
                fprintf(stderr, "%s: local host %s\n", prog,
                       inet_ntoa(localhost));
            }
#endif
        }
#endif

#ifdef LINUX_FPU_FIX
    __setfpucw(_FPU_IEEE);            /* Mask floating point interrupts to
                                         enable standard IEEE processing.
                                         Current libc releases do this by
                                         default, so this is needed only on
                                         older libraries. */
#endif

        /* Initialise GSM decoding. */

        gsmh = gsm_create();

#ifdef ENABLE_LPC        
        /* Initialise LPC decoding. */

        if (!lpc_start()) {
            fprintf(stderr, "Cannot allocate LPC decoding memory.\n");
            return 1;
        }
        lpc10init();
        
        /* Initialise CELP decoding. */
        
        celp_init(FALSE);
#endif        

        /* Find assigned port value and print it. */

        length = sizeof(name);
        if (getsockname(sock, (struct sockaddr *) &name, &length) < 0) {
            perror("getting socket name");
            return 1;
        }
#ifdef SHOW_SOCKET
        fprintf(stderr, "%s: socket port #%d\n", prog, ntohs(name.sin_port));
#endif
    }

    /*  See if the user has specified a face file.  If so,
        try to open it.  */

    signal(SIGHUP, exiting);          /* Set signal to handle termination */
    signal(SIGINT, exiting);          /* Set signal to handle termination */
    signal(SIGTERM, exiting);         /* Set signal to handle termination */
    signal(SIGALRM, release);         /* Set signal to handle timeout */
    signal(SIGCHLD, childterm);       /* Set signal to handle face viewer termination */
    windtimer();                      /* Set periodic status update  */

#ifdef NAT_LAUNCH
    /* Launch sfmike if -NAThost was specified. */
/*    if (xnat_conspec) {
        xnat_active_sfmike_pid = launch_sfmike(NULL, mikeopts);
    }*/
#endif

    /* Read from the socket. */

    while (TRUE) {
        fd_set fdset;
        int wsock, control, busyreject;

        FD_ZERO(&fdset);
        FD_SET(sock, &fdset);
        FD_SET(ssock, &fdset);
        
        /*  Wait here for a packet to arrive on either the control or
            audio socket.  */
            
        if (select(FD_SETSIZE, &fdset, NULL, NULL, NULL) <= 0) {
            continue;
        }
        wsock = FD_ISSET(sock, &fdset) ? sock :
                    (FD_ISSET(ssock, &fdset) ? ssock : -1);
        if (wsock < 0) {
            continue;
        }
        
        /*  A packet appears to have arrived.  Read it from the proper socket.  */
        
        control = wsock == ssock;
        fromlen = sizeof(from);
        if ((rll = recvfrom(wsock, (char *) &sb, sizeof sb, 0, (struct sockaddr *) &from, &fromlen)) < 0) {
            if (errno != EINTR) {
                perror(!control ? "receiving data packet" :
                                  "receiving control packet");
            }
            continue;
        }
        //if (wsock == ssock) dbg("got control packet\n");
#ifdef HEXDUMP
        if (hexdump) {
            fprintf(stderr, "speaker: %d bytes read from %s socket.\n", rll,
                    control ? "control" : "data");
            xd(stderr, &sb, rll, TRUE);
        }
#endif

        /* See if this connection is active.  If not, initialise a new
           connection. */

        

        busyreject = FALSE;
        newconn = FALSE;
        c = conn;
        while (c != NULL) {
            if (memcmp(&from.sin_addr, &(c->con_addr),
                       sizeof(struct in_addr)) == 0) {
                break;
            }
            c = c->con_next;
        }
        
        /*  If we didn't find a connection from the sender's address
            already active, open a new one and insert it in the list
            of connections.  */
            
        if (c == NULL) {
            c = (struct connection *) malloc(sizeof(struct connection));
            if (c != NULL) {
                struct hostent *h;

                newconn = TRUE;
                c->con_next = conn;
                c->pgpkey[0] = FALSE;
                bzero(c->keymd5, 16);
                c->con_uname[0] = c->con_email[0] = 0;
                conn = c;
                bcopy(&from.sin_addr, &(c->con_addr),
                    sizeof(struct in_addr));
                h = gethostbyaddr((char *) &from.sin_addr, sizeof(struct in_addr),
                                  AF_INET);
                if (h == NULL) {
                    strcpy(c->con_hostname, inet_ntoa(from.sin_addr));
                } else {
                    strcpy(c->con_hostname, h->h_name);
                }
#ifdef NAT_LAUNCH
                if (control) {
                    c->con_ports.cport = ntohs(from.sin_port);
                    c->con_ports.dport = 0;
                } else {
                    c->con_compression = sb.compression;
                    c->con_ports.dport = ntohs(from.sin_port);
                    c->con_ports.cport = 0;
                }
                c->sfmike_pid = (xnat_conspec ? xnat_active_sfmike_pid : 0);
#endif
            }
        } else if (c->con_timeout == -1) {
            newconn = TRUE;
        }

        /* Initialise fields in connection.  Only fields which need to
           be reinitialised when a previously idle host resumes activity
           need be set here. */

        if (newconn) {
#ifdef NAT_LAUNCH
            c->sfmike_pid = 0;
#endif
            c->con_compmodes = -1;
            c->con_protocol = PROTOCOL_UNKNOWN;
            c->con_rseq = -1;
            c->con_reply_current = FALSE;
            c->con_busy = 0;
            bcopy("\221\007\311\201", c->con_session, 4);
#ifdef ENABLE_LPC            
            lpc_init(&c->lpcc);
#endif            
            busyreject = isBusy();
        }

        if (c != NULL) {
            /* Reset connection timeout. */
            c->con_timeout = 0;

            /* If we're rejecting this with a busy signal, start
               the busy timeout running and dispatch the busy
               signal to the host. */

            if (busyreject && c->con_busy == 0) {
                pid_t cpid;

                c->con_busy = busytimeout;

                if (audiok) {
                    soundterm();
                    audiok = FALSE;
                    if (Debug) {
                         fprintf(stderr, "speaker: releasing audio before busy signal fork().\n");
                    }
                }
                cpid = fork();
                if (cpid == 0) {
                    char s[256];

                    signal(SIGHUP, SIG_DFL);
                    signal(SIGINT, SIG_DFL);
                    signal(SIGTERM, SIG_DFL);
                    signal(SIGALRM, SIG_DFL);
                    signal(SIGCHLD, SIG_DFL);

                    /* Now we need to close any shared resources
                       that might have been inherited from the parent
                       process to avoid their being locked up for the
                       duration of the child process execution. */

                    close(sock);
                    if (record != NULL) {
                        fclose(record);
                    }
                    sprintf(s, busysignal, inet_ntoa(from.sin_addr));
                    if (Debug) {
                        fprintf(stderr, "speaker: sending busy signal with %s\n", s);
                    }
                    system(s);
                    exit(0);
                } else if (cpid == (pid_t) -1) {
                    perror("creating busy signal process");
                }
            }

            if (newconn) {
                if (showhosts) {
                    fprintf(stderr, "speaker: %s %s %s\n",etime(),
                            c->con_hostname,
                            busyreject ? "sending busy signal" : "connect");
                }
            }
            if (busyreject) {
                continue;
            }

            
#ifdef NAT_LAUNCH
            if (xnat_passive && (c->sfmike_pid == 0)) {
                if (control && (c->con_ports.cport == 0)) {
                    c->con_ports.cport = ntohs(from.sin_port);
                }
                if ((!control) && (c->con_ports.dport == 0)) {
                    c->con_ports.dport = ntohs(from.sin_port);
                    c->con_compression = sb.compression;
                }
/*                if ((c->con_ports.dport > 0) && c->con_ports.cport) {
                    c->sfmike_pid = launch_sfmike(c, mikeopts);
                }*/
            }
#endif
        } else {
            continue;
        }

        /* If we've sent the connection a busy signal, ignore all
           packets from it until the busy timeout expires. */

        if (c->con_busy > 0) {
            if (Debug) {
                fprintf(stderr, "speaker: busy--discarding packet from %s.\n", c->con_hostname);
            }
            continue;
        }

        wasrtp = FALSE;


        /* If this packet arrived on the session control port, dispatch
           it to the appropriate handler for its protocol. */

        if (control) {
            short protocol = PROTOCOL_VATRTP_CRYPT;
            unsigned char *p = (unsigned char *) &sb;
            unsigned char *apkt;
            int proto = (p[0] >> 6) & 3;

            if (proto == 0) {         /* VAT */
                /* To avoid spoofing by bad encryption keys, require
                   a proper ID message be seen before we'll flip into
                   VAT protocol. */
                if (((p[1] == 1) || (p[1] == 3)) ||
                    ((c->con_protocol == PROTOCOL_VAT) && (p[1] == 2))) {
                    protocol = PROTOCOL_VAT;
                    bcopy(p + 2, c->con_session, 2);  /* Save conference ID */

                    if (p[1] == 1 && showhosts) {
                        char uname[256];
                        int unl = rll - 4;

                        if (unl > ((sizeof uname) - 1)) {
                            unl = (sizeof uname) - 1;
                        }
                        bcopy(p + 4, uname, unl);
                        uname[unl] = 0;
                        if (strcmp(uname, c->con_uname) != 0) {
                            strcpy(c->con_uname, uname);
                            fprintf(stderr, "speaker: %s sending from %s.\n", c->con_uname, c->con_hostname);
                        }
                    }

                    /* Handling of VAT IDLIST could be a lot more elegant
                       than this. */

                    if (p[1] == 3 && showhosts) {
                        char *uname;

                        uname = (char *) malloc(rll);
                        if (uname != NULL) {
                            unsigned char *bp = p, *ep = p + rll;
                            int i = bp[4];

                            bp += 8;
                            uname[0] = 0;
                            *ep = 0;
                            while (--i >= 0 && bp < ep) {
                                bp += 4;
                                strcat(uname, "\t");
                                strcat(uname, (char *) bp);
                                while (isspace(uname[strlen(uname) - 1])) {
                                    uname[strlen(uname) - 1] = 0;
                                }
                                strcat(uname, "\n");
                                bp += (strlen((char *) bp) + 3) & ~3;
                            }
                            if (strncmp(uname, c->con_uname, (sizeof c->con_uname - 1)) != 0) {
                                strncpy(c->con_uname, uname, sizeof c->con_uname);
                                if (strlen(uname) > ((sizeof c->con_uname) - 1)) {
                                    c->con_uname[((sizeof c->con_uname) - 1)] = 0;
                                }
                                fprintf(stderr, "speaker: now in conference at %s:\n%s", c->con_hostname, uname);
                            }
                            free(uname);
                        }
                    }

                    /* If it's a DONE packet, reset protocol to unknown. */

                    if (p[1] == 2) {
                        c->con_protocol = protocol = PROTOCOL_UNKNOWN;
                        c->con_timeout = hosttimeout - 1;
                        c->con_uname[0] = c->con_email[0] = 0;
                        if (showhosts) {
                            fprintf(stderr, "speaker: %s VAT connection closed.\n", c->con_hostname);
                        }
                    }
                }

            } else if (proto == RTP_VERSION || proto == 1) { /* RTP */
                if (isValidRTCPpacket((unsigned char *) &sb, rll)) {
                    protocol = (proto == 1) ? PROTOCOL_SPEAKFREE : PROTOCOL_RTP;
                    bcopy(p + 4, c->con_session, 4);  /* Save SSRC */

                    /* If it's a BYE packet, reset protocol to unknown. */

                    if (isRTCPByepacket((unsigned char *) &sb, rll)) {
                        c->con_protocol = protocol = PROTOCOL_UNKNOWN;
                        c->con_timeout = hosttimeout - 1;
                        c->con_uname[0] = c->con_email[0] = 0;
                        if (showhosts) {
                            fprintf(stderr, "speaker: %s %s connection closed.\n", c->con_hostname,
                                    proto == 1 ? "Speak Freely" : "RTP");
                        }

                    /* If it's a text chat message, print it. */

                    } else if (isRTCPAPPpacket((unsigned char *) &sb, rll,
                                    RTCP_APP_TEXT_CHAT, &apkt) && apkt != NULL) {
                        char *ident = c->con_hostname;

                        /* To identify the sender, get successively more
                           personal depending on the information we have at
                           hand, working down from hostname (which may just
                           be an IP address if we couldn't resolve the host,
                           through E-mail address, to user name. */

                        if (c->con_email[0] != 0) {
                            ident = c->con_email;
                        }
                        if (c->con_uname[0] != 0) {
                            ident = c->con_uname;
                        }

                        printf("%s: %s\n", ident, (char *) (apkt + 12));

                    /* Otherwise, it's presumably an SDES, from which we
                       should update the user identity information for the
                       connection. */

                    } else {
                        struct rtcp_sdes_request rp;

                        rp.nitems = 4;
                        rp.item[0].r_item = RTCP_SDES_CNAME;
                        rp.item[1].r_item = RTCP_SDES_NAME;
                        rp.item[2].r_item = RTCP_SDES_EMAIL;
                        rp.item[3].r_item = RTCP_SDES_TOOL;
                        if (parseSDES((unsigned char *) &sb, &rp)) {
                            char uname[256], email[256];

                            uname[0] = email[0] = 0;
                            if (rp.item[1].r_text != NULL) {
                                copySDESitem(rp.item[1].r_text, uname);
                                if (rp.item[2].r_text != NULL) {
                                    copySDESitem(rp.item[2].r_text, email);
                                } else if (rp.item[2].r_text != NULL) {
                                    copySDESitem(rp.item[0].r_text, email);
                                }
                            } else if (rp.item[2].r_text != NULL) {
                                copySDESitem(rp.item[2].r_text, uname);
                            } else if (rp.item[0].r_text != NULL) {
                                copySDESitem(rp.item[0].r_text, uname);
                            }
                            if (strcmp(uname, c->con_uname) != 0 ||
                                strcmp(email, c->con_email) != 0) {
                                strcpy(c->con_uname, uname);
                                strcpy(c->con_email, email);
                                if (showhosts && uname[0]) {
                                    fprintf(stderr, "speaker: %s", uname);
                                    if (email[0]) {
                                      fprintf(stderr, " (%s)", email);
                                    }
                                    fprintf(stderr, " sending from %s.\n",
                                            c->con_hostname);
                                }
                            }
                        }
                    }
                } else {
                    if (Debug) {
                        fprintf(stderr, "Invalid RTCP packet received.\n");
                    }
                }
            } else {
                if (Debug) {
                    fprintf(stderr, "Bogus protocol 3 control message.\n");
                }
            }

            /* If protocol changed, update in connection and, if appropriate,
               update the reply command. */

            if (protocol != c->con_protocol) {
                static char *pname[] = {
                    "Speak Freely",
                    "VAT",
                    "RTP",
                    "VAT/RTP encrypted",
                    "Unknown"
                };

                c->con_protocol = protocol;
                if (showhosts) {
                    fprintf(stderr, "speaker: %s sending in %s protocol.\n", c->con_hostname, pname[protocol]);
                }
                c->con_reply_current = FALSE;
            }
            continue;
        }

        /* If this message is tagged with our Speak Freely protocol
           bit, force protocol back to Speak Freely.  This allows us
           to switch back to Speak Freely after receiving packets in
           VAT.  We can still get confused if we receive a packet from
           an older version of Speak Freely that doesn't tag. */

        if (c->con_protocol == PROTOCOL_VAT ||
            c->con_protocol == PROTOCOL_VATRTP_CRYPT) {
            unsigned char *p = (unsigned char *) &sb;

            if (((p[0] >> 6) & 3) == 1) {
                c->con_protocol = PROTOCOL_SPEAKFREE;
            }
        }

        /* If this is a VAT packet, translate it into a sound buffer. */

        if (((c->con_protocol == PROTOCOL_VAT)) &&
            (bcmp(((unsigned char *) &sb) + 2, c->con_session, 2) == 0) &&
            isvat((unsigned char *) &sb, rll)) {
            if (sb.buffer.buffer_len == 0) {
                if (Debug) {
                    fprintf(stderr, "Ignoring unparseable VAT packet.\n");
                }
                continue;
            }
            wasrtp = TRUE;

        /* If this is an RTP packet, transmogrify it into a sound
           buffer we can understand. */

        } else if ((c->con_protocol == PROTOCOL_RTP) &&
                 (bcmp(((unsigned char *) &sb) + 8, c->con_session, 4) == 0) &&
                 isrtp((unsigned char *) &sb, rll)) {
            if (sb.buffer.buffer_len == 0) {
                if (Debug) {
                    fprintf(stderr, "Ignoring unparseable RTP packet.\n");
                }
                continue;
            }
            wasrtp = TRUE;
        }

        //dump_packet(&sb);
        if (!wasrtp) {
            long xbl;

            /* Convert relevant fields from network to host
               byte order, if necessary. */

            sb.compression = ntohl(sb.compression);
            sb.buffer.buffer_len = ntohl(sb.buffer.buffer_len);

            if (sb.compression & fCompRobust) {
                int aseq = (sb.buffer.buffer_len >> 24) & 0xFF, pdist, mdist, cdist;

                /* Robust mode sends multiple copies of each sound buffer
                   with a sequence number modulo 256 hidden in the most
                   significant byte of the buffer length field.  If
                   multipath has caused packets to arrive out of order,
                   we want do discard any packets prior one we've played;
                   otherwise we'd shuffle the sound with the result being
                   gibberish.  The fact that the sequence number is modular
                   complicates this test a tad since the sequence numbers
                   wrap around at 255.  Consequently, we compute the
                   distance in modular space between the sequence number
                   in the packet we've just received and that of the last
                   packet we played.  */

                if (aseq >= c->con_rseq) {
                    pdist = aseq - c->con_rseq;
                    mdist = (c->con_rseq + 256) - aseq;
                } else {
                    mdist = c->con_rseq - aseq;
                    pdist = (aseq + 256) - c->con_rseq;
                }
                cdist = (pdist <= mdist) ? pdist : -mdist;

                if ((cdist <= 0) && (cdist > -16)) {
                    if (Debug) {
                        fprintf(stderr, "Robust REJECTING   seq %3d   last seq %3d   distance %4d\n",
                            aseq, c->con_rseq, cdist);
                    }
                    continue;
                }
                if (Debug) {
                    fprintf(stderr, "Robust PLAYING     seq %3d   last seq %3d   distance %4d\n",
                        aseq, c->con_rseq, cdist);
                }
                c->con_rseq = aseq;
                sb.buffer.buffer_len &= 0xFFFFFFL;
            }

            /* Now if this is a valid Speak Freely packet (as
               opposed to a VAT packet masquerading as one, or
               an encrypted VAT or RTP packet we don't have the
               proper key to decode), the length received from the
               socket will exactly equal the buffer length plus
               the size of the header.  This is a reasonably
               good validity check, well worth it considering the
               horrors treating undecipherable garbage as a sound
               buffer could inflict on us. */

            xbl = sb.buffer.buffer_len + (sizeof(struct soundbuf) - BUFL);

            /* If this packet is encrypted with an algorithm which requires
               padding the packet to an 8- or 16-byte boundary, adjust the actual
               content length to account for the padding. */

            if (1) {
                if ((sb.compression & (fEncDES | fEncIDEA | fEncBF | fEncPGP)) != 0) {
                    xbl = ((sb.buffer.buffer_len + 7) & (~7)) +
                          (sizeof(struct soundbuf) - BUFL);
                }

                if ((sb.compression & fEncAES) != 0) {
                    xbl = ((sb.buffer.buffer_len + 15) & (~15)) +
                          (sizeof(struct soundbuf) - BUFL);
                }

                /* If packet is compressed with LPC-10 or CELP, compensate for "packet
                   stuffing". */

                if ((sb.compression & (fCompLPC10 | fCompCELP)) &&
                    (sb.buffer.buffer_len >= 16)) {
                    xbl -= 16;
                }
                if (xbl != rll) {
                    if (Debug) {
                        fprintf(stderr,
                                "speaker: Sound buffer from %s length %ld doesn't match %d byte packet.  Hdr=%08lX\n",
                                c->con_hostname, xbl, rll, sb.compression);
                    }
                    if (showhosts && c->con_protocol != PROTOCOL_UNKNOWN) {
                        fprintf(stderr, "speaker: %s sending in unknown protocol or encryption.\n",
                                c->con_hostname);
                    }
                    c->con_protocol = PROTOCOL_UNKNOWN;
                    continue;
                }
            } 
            /* The only packets which remain are half-duplex mute requests,
               and these contain no payload. */

            /* It does appear to be a genuine Speak Freely sound
               buffer.  On that basis, set the protocol to Speak Freely
               even if the buffer isn't explicitly tagged. */

            if (c->con_protocol != PROTOCOL_SPEAKFREE) {
                c->con_protocol = PROTOCOL_SPEAKFREE;
                if (showhosts) {
                    fprintf(stderr, "speaker: %s sending in Speak Freely protocol.\n", c->con_hostname);
                }
                c->con_reply_current = FALSE;
            }
        }
        
        if ((replycmdexe != NULL && !c->con_reply_current)) {
            char *rcmd = StrReplace(replycmdexe, "%s", c->con_hostname);

            if (rcmd != NULL) {
                if (Debug) {
                    fprintf(stderr, "Executing: %s\n", rcmd);
                }
                system(rcmd);
                c->con_reply_current = TRUE;
                if (rcmd != replycmdexe) {
                    free(rcmd);
                }
            }
        }

        /* If this is the first sound packet we've seen from this host 
           in this protocol, create (or update) the reply command file. */

        if (replyfile != NULL && !c->con_reply_current) {

            /* If a reply file is requested, create an executable
               shell script to reply to the host that just connected
               using either the default command or whatever command
               the user specifies on the -A option. */

            FILE *rfp = fopen(replyfile, "w");
            static char *popt[] = {
                "",
                "-vat ",
                "-rtp ",
                "-rtp ",
                ""
            };

            if (rfp != NULL) {
                fprintf(rfp, "#! /bin/sh\n");
                /* $* allows specifying options on reply file call. */
                fprintf(rfp, "%s %s$* %s\n", replycmd,
                              popt[c->con_protocol], c->con_hostname);
                /* Make the file executable. */
                fchmod(fileno(rfp), 0755);
                fclose(rfp);
            }
            c->con_reply_current = TRUE;
        }

#ifdef HALF_DUPLEX

        /* If this is a half duplex mute request, immediately release
           the audio output device if we have it.  We verify the
           sender's address and accept mute requests only frome
           localhost; nobody else has any business telling us
           to shut up! */


        if (sb.compression & fHalfDuplex) {
#ifdef HDX_DEBUG
            if (Debug) {
                static char *hdxreq[4] = { "Bogus(0)", "Mute", "Resume",
                                           "Bogus(3)" };

                fprintf(stderr, "speaker: half-duplex %s request from %s.\n",
                    hdxreq[sb.compression & 3], inet_ntoa(from.sin_addr));
            }
#endif
            if (memcmp(&from.sin_addr, &localhost, sizeof localhost) == 0) {
                if (sb.compression & fHalfDuplexMute) {
                    if (audiok) {
                        soundterm();
                        audiok = FALSE;
                        if (Debug) {
                            fprintf(stderr, "speaker: half-duplex releasing audio output.\n");
                        }
                    }
                    halfDuplexMuted = TRUE;
                } else if (sb.compression & fHalfDuplexResume) {
                    halfDuplexMuted = FALSE;
                }
            }
            continue;                 /* Done with packet */
        }
#endif


        /* If the packet requests loop-back, immediately dispatch it
           back to the host who sent it to us.  To prevent an infinite
           loop-back cycle, we clear the loop-back bit in the header
           before sending the message.  We leave the host of origin
           unchanged, allowing the sender to identify the packet as
           one he originated. */

        if (sb.compression & fLoopBack) {
            bcopy((char *) &(from.sin_addr), (char *) &(name.sin_addr),
                sizeof(struct in_addr));
            sb.compression &= ~fLoopBack;    /* Prevent infinite loopback */

            sb.compression = htonl(sb.compression);
            sb.buffer.buffer_len = htonl(sb.buffer.buffer_len);
            if (sendto(sock, (char *) &sb, rll,
                0, (struct sockaddr *) &(name), sizeof name) < 0) {
                perror("sending datagram message");
            }
            sb.compression = ntohl(sb.compression);
            sb.buffer.buffer_len = ntohl(sb.buffer.buffer_len);
        }

#ifdef HALF_DUPLEX

        /* If we're muted by a transmission in progress on half-duplex
           audio hardware, this is the end of line for this sound
           packet. */

        if (halfDuplexMuted) {
            if (Debug) {
                fprintf(stderr, "speaker: %s packet lost by half-duplex muting.\n",
                                c->con_hostname);
            }
            continue;
        }
#endif

        /* If this packet has been "stuffed" for maximum efficiency,
           un-stuff it at this point. */

        if ((sb.compression & (fCompLPC10 | fCompCELP)) &&
            (sb.buffer.buffer_len >= 16)) {
            bcopy(sb.sendinghost, (char *) &sb + rll,
                  sizeof sb.sendinghost);
            rll += sizeof sb.sendinghost;
        }

        {
            playbuffer(&sb, c);
        }
    }
#ifdef MEANS_OF_EXIT
    close(sock);
    gsm_destroy(gsmh);
    lpc_end();
    exiting();
    if (record != NULL) {
        fclose(record);
    }
    return 0;
#endif
}

int speaker_stop(void){
    if (!speaker_pid) return 0;

	kill(speaker_pid, SIGTERM);

	dbg("waiting for speaker pid %d\n", speaker_pid);
	waitpid(speaker_pid, NULL, 0);
	dbg("speaker pid %d terminated\n", speaker_pid);
	speaker_pid = 0;
	return(0);
}
